# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compiler_opt.rl.agent_config."""

import gin
import tensorflow as tf
from tf_agents.agents.behavioral_cloning import behavioral_cloning_agent
from tf_agents.agents.dqn import dqn_agent
from tf_agents.agents.ppo import ppo_agent
from tf_agents.networks import actor_distribution_network
from tf_agents.networks import q_network
from tf_agents.specs import tensor_spec
from tf_agents.trajectories import time_step

from compiler_opt.rl import agent_config


def _observation_processing_layer(obs_spec):
  del obs_spec

  def identity(obs):
    return tf.expand_dims(tf.cast(obs, tf.float32), -1)

  return tf.keras.layers.Lambda(identity)


class AgentCreatorsTest(tf.test.TestCase):

  def setUp(self):
    observation_spec = tf.TensorSpec(
        dtype=tf.int64, shape=(), name='callee_users')
    self._time_step_spec = time_step.time_step_spec(observation_spec)
    self._action_spec = tensor_spec.BoundedTensorSpec(
        dtype=tf.int64,
        shape=(),
        minimum=0,
        maximum=1,
        name='inlining_decision')
    super().setUp()

  def test_create_behavioral_cloning_agent(self):
    gin.bind_parameter('create_agent.policy_network', q_network.QNetwork)
    gin.bind_parameter('BehavioralCloningAgent.optimizer',
                       tf.compat.v1.train.AdamOptimizer())
    tf_agent = agent_config.create_agent(
        agent_config.BCAgentConfig(
            time_step_spec=self._time_step_spec, action_spec=self._action_spec),
        preprocessing_layer_creator=_observation_processing_layer)
    self.assertIsInstance(tf_agent,
                          behavioral_cloning_agent.BehavioralCloningAgent)

  def test_create_dqn_agent(self):
    gin.bind_parameter('create_agent.policy_network', q_network.QNetwork)
    gin.bind_parameter('DqnAgent.optimizer', tf.compat.v1.train.AdamOptimizer())
    tf_agent = agent_config.create_agent(
        agent_config.DQNAgentConfig(
            time_step_spec=self._time_step_spec, action_spec=self._action_spec),
        preprocessing_layer_creator=_observation_processing_layer)
    self.assertIsInstance(tf_agent, dqn_agent.DqnAgent)

  def test_create_ppo_agent(self):
    gin.bind_parameter('create_agent.policy_network',
                       actor_distribution_network.ActorDistributionNetwork)
    gin.bind_parameter('PPOAgent.optimizer', tf.compat.v1.train.AdamOptimizer())
    tf_agent = agent_config.create_agent(
        agent_config.PPOAgentConfig(
            time_step_spec=self._time_step_spec, action_spec=self._action_spec),
        preprocessing_layer_creator=_observation_processing_layer)
    self.assertIsInstance(tf_agent, ppo_agent.PPOAgent)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for policy_utils."""

import os

from absl.testing import absltest
import numpy as np
import tensorflow as tf
from tf_agents.networks import actor_distribution_network
from tf_agents.policies import actor_policy
from tf_agents.policies import tf_policy

from compiler_opt.es import policy_utils
from compiler_opt.rl import inlining
from compiler_opt.rl import policy_saver
from compiler_opt.rl import regalloc
from compiler_opt.rl import registry
from compiler_opt.rl.inlining import config as inlining_config
from compiler_opt.rl.regalloc import config as regalloc_config
from compiler_opt.rl.regalloc import regalloc_network


class ConfigTest(absltest.TestCase):

  # TODO(abenalaast): Issue #280
  def test_inlining_config(self):
    problem_config = registry.get_configuration(
        implementation=inlining.InliningConfig)
    time_step_spec, action_spec = problem_config.get_signature_spec()
    quantile_file_dir = os.path.join('compiler_opt', 'rl', 'inlining', 'vocab')
    creator = inlining_config.get_observation_processing_layer_creator(
        quantile_file_dir=quantile_file_dir,
        with_sqrt=False,
        with_z_score_normalization=False)
    layers = tf.nest.map_structure(creator, time_step_spec.observation)

    actor_network = actor_distribution_network.ActorDistributionNetwork(
        input_tensor_spec=time_step_spec.observation,
        output_tensor_spec=action_spec,
        preprocessing_layers=layers,
        preprocessing_combiner=tf.keras.layers.Concatenate(),
        fc_layer_params=(64, 64, 64, 64),
        dropout_layer_params=None,
        activation_fn=tf.keras.activations.relu)

    policy = actor_policy.ActorPolicy(
        time_step_spec=time_step_spec,
        action_spec=action_spec,
        actor_network=actor_network)

    self.assertIsNotNone(policy)
    self.assertIsInstance(
        policy._actor_network,  # pylint: disable=protected-access
        actor_distribution_network.ActorDistributionNetwork)

  # TODO(abenalaast): Issue #280
  def test_regalloc_config(self):
    problem_config = registry.get_configuration(
        implementation=regalloc.RegallocEvictionConfig)
    time_step_spec, action_spec = problem_config.get_signature_spec()
    quantile_file_dir = os.path.join('compiler_opt', 'rl', 'regalloc', 'vocab')
    creator = regalloc_config.get_observation_processing_layer_creator(
        quantile_file_dir=quantile_file_dir,
        with_sqrt=False,
        with_z_score_normalization=False)
    layers = tf.nest.map_structure(creator, time_step_spec.observation)

    actor_network = regalloc_network.RegAllocNetwork(
        input_tensor_spec=time_step_spec.observation,
        output_tensor_spec=action_spec,
        preprocessing_layers=layers,
        preprocessing_combiner=tf.keras.layers.Concatenate(),
        fc_layer_params=(64, 64, 64, 64),
        dropout_layer_params=None,
        activation_fn=tf.keras.activations.relu)

    policy = actor_policy.ActorPolicy(
        time_step_spec=time_step_spec,
        action_spec=action_spec,
        actor_network=actor_network)

    self.assertIsNotNone(policy)
    self.assertIsInstance(
        policy._actor_network,  # pylint: disable=protected-access
        regalloc_network.RegAllocNetwork)


class VectorTest(absltest.TestCase):

  expected_variable_shapes = [(71, 64), (64), (64, 64), (64), (64, 64), (64),
                              (64, 64), (64), (64, 2), (2)]
  expected_length_of_a_perturbation = sum(
      np.prod(shape) for shape in expected_variable_shapes)
  params = np.arange(expected_length_of_a_perturbation, dtype=np.float32)
  POLICY_NAME = 'test_policy_name'

  # TODO(abenalaast): Issue #280
  def test_set_vectorized_parameters_for_policy(self):
    # create a policy
    problem_config = registry.get_configuration(
        implementation=inlining.InliningConfig)
    time_step_spec, action_spec = problem_config.get_signature_spec()
    quantile_file_dir = os.path.join('compiler_opt', 'rl', 'inlining', 'vocab')
    creator = inlining_config.get_observation_processing_layer_creator(
        quantile_file_dir=quantile_file_dir,
        with_sqrt=False,
        with_z_score_normalization=False)
    layers = tf.nest.map_structure(creator, time_step_spec.observation)

    actor_network = actor_distribution_network.ActorDistributionNetwork(
        input_tensor_spec=time_step_spec.observation,
        output_tensor_spec=action_spec,
        preprocessing_layers=layers,
        preprocessing_combiner=tf.keras.layers.Concatenate(),
        fc_layer_params=(64, 64, 64, 64),
        dropout_layer_params=None,
        activation_fn=tf.keras.activations.relu)

    policy = actor_policy.ActorPolicy(
        time_step_spec=time_step_spec,
        action_spec=action_spec,
        actor_network=actor_network)

    # save the policy
    saver = policy_saver.PolicySaver({VectorTest.POLICY_NAME: policy})
    testing_path = self.create_tempdir()
    policy_save_path = os.path.join(testing_path, 'temp_output', 'policy')
    saver.save(policy_save_path)

    # set the values of the policy variables
    policy_utils.set_vectorized_parameters_for_policy(policy, VectorTest.params)
    # iterate through variables and check their shapes and values
    # deep copy params in order to destructively iterate over values
    expected_values = [*VectorTest.params]
    for i, variable in enumerate(policy.variables()):  # pylint: disable=not-callable
      self.assertEqual(variable.shape, VectorTest.expected_variable_shapes[i])
      variable_values = variable.numpy().flatten()
      np.testing.assert_array_almost_equal(
          expected_values[:len(variable_values)], variable_values)
      expected_values = expected_values[len(variable_values):]
    # all values in the copy should have been removed at this point
    self.assertEmpty(expected_values)

    # get saved model to test a loaded policy
    load_path = os.path.join(policy_save_path, VectorTest.POLICY_NAME)
    sm = tf.saved_model.load(load_path)
    self.assertNotIsInstance(sm, tf_policy.TFPolicy)
    policy_utils.set_vectorized_parameters_for_policy(sm, VectorTest.params)
    # deep copy params in order to destructively iterate over values
    expected_values = [*VectorTest.params]
    for i, variable in enumerate(sm.model_variables):
      self.assertEqual(variable.shape, VectorTest.expected_variable_shapes[i])
      variable_values = variable.numpy().flatten()
      np.testing.assert_array_almost_equal(
          expected_values[:len(variable_values)], variable_values)
      expected_values = expected_values[len(variable_values):]
    # all values in the copy should have been removed at this point
    self.assertEmpty(expected_values)

  # TODO(abenalaast): Issue #280
  def test_get_vectorized_parameters_from_policy(self):
    # create a policy
    problem_config = registry.get_configuration(
        implementation=inlining.InliningConfig)
    time_step_spec, action_spec = problem_config.get_signature_spec()
    quantile_file_dir = os.path.join('compiler_opt', 'rl', 'inlining', 'vocab')
    creator = inlining_config.get_observation_processing_layer_creator(
        quantile_file_dir=quantile_file_dir,
        with_sqrt=False,
        with_z_score_normalization=False)
    layers = tf.nest.map_structure(creator, time_step_spec.observation)

    actor_network = actor_distribution_network.ActorDistributionNetwork(
        input_tensor_spec=time_step_spec.observation,
        output_tensor_spec=action_spec,
        preprocessing_layers=layers,
        preprocessing_combiner=tf.keras.layers.Concatenate(),
        fc_layer_params=(64, 64, 64, 64),
        dropout_layer_params=None,
        activation_fn=tf.keras.activations.relu)

    policy = actor_policy.ActorPolicy(
        time_step_spec=time_step_spec,
        action_spec=action_spec,
        actor_network=actor_network)

    # save the policy
    saver = policy_saver.PolicySaver({VectorTest.POLICY_NAME: policy})
    testing_path = self.create_tempdir()
    policy_save_path = os.path.join(testing_path, 'temp_output', 'policy')
    saver.save(policy_save_path)

    # functionality verified in previous test
    policy_utils.set_vectorized_parameters_for_policy(policy, VectorTest.params)
    # vectorize and check if the outcome is the same as the start
    output = policy_utils.get_vectorized_parameters_from_policy(policy)
    np.testing.assert_array_almost_equal(output, VectorTest.params)

    # get saved model to test a loaded policy
    load_path = os.path.join(policy_save_path, VectorTest.POLICY_NAME)
    sm = tf.saved_model.load(load_path)
    self.assertNotIsInstance(sm, tf_policy.TFPolicy)
    policy_utils.set_vectorized_parameters_for_policy(sm, VectorTest.params)
    # vectorize and check if the outcome is the same as the start
    output = policy_utils.get_vectorized_parameters_from_policy(sm)
    np.testing.assert_array_almost_equal(output, VectorTest.params)

  # TODO(abenalaast): Issue #280
  def test_tfpolicy_and_loaded_policy_produce_same_variable_order(self):
    # create a policy
    problem_config = registry.get_configuration(
        implementation=inlining.InliningConfig)
    time_step_spec, action_spec = problem_config.get_signature_spec()
    quantile_file_dir = os.path.join('compiler_opt', 'rl', 'inlining', 'vocab')
    creator = inlining_config.get_observation_processing_layer_creator(
        quantile_file_dir=quantile_file_dir,
        with_sqrt=False,
        with_z_score_normalization=False)
    layers = tf.nest.map_structure(creator, time_step_spec.observation)

    actor_network = actor_distribution_network.ActorDistributionNetwork(
        input_tensor_spec=time_step_spec.observation,
        output_tensor_spec=action_spec,
        preprocessing_layers=layers,
        preprocessing_combiner=tf.keras.layers.Concatenate(),
        fc_layer_params=(64, 64, 64, 64),
        dropout_layer_params=None,
        activation_fn=tf.keras.activations.relu)

    policy = actor_policy.ActorPolicy(
        time_step_spec=time_step_spec,
        action_spec=action_spec,
        actor_network=actor_network)

    # save the policy
    saver = policy_saver.PolicySaver({VectorTest.POLICY_NAME: policy})
    testing_path = self.create_tempdir()
    policy_save_path = os.path.join(testing_path, 'temp_output', 'policy')
    saver.save(policy_save_path)

    # set the values of the variables
    policy_utils.set_vectorized_parameters_for_policy(policy, VectorTest.params)
    # save the changes
    saver.save(policy_save_path)
    # vectorize the tfpolicy
    tf_params = policy_utils.get_vectorized_parameters_from_policy(policy)

    # get loaded policy
    load_path = os.path.join(policy_save_path, VectorTest.POLICY_NAME)
    sm = tf.saved_model.load(load_path)
    # vectorize the loaded policy
    loaded_params = policy_utils.get_vectorized_parameters_from_policy(sm)

    # assert that they result in the same order of values
    np.testing.assert_array_almost_equal(tf_params, loaded_params)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Lists gtests in a JSON format
This script takes in a gtest executable and outputs all of the tests
contained within that gtest executable to the command line separated by
line breaks or in a JSON format.

Usage:
PYTHONPATH=$PYTHONPATH:. python3 \
compiler_opt/benchmark/list_gtests.py \
  --gtest_executable=/path/to/executable \
  --output_type=json \
  --output_file=/path/to/output

The gtest_executable flag is required. The output_type by default is set
to JSON. If you don't provide an output file path, the output will just
be printed to the terminal (fully compatible with shell pipes).
"""

import json
import sys
import os

from absl import app
from absl import flags
from absl import logging

from compiler_opt.benchmark import gtest_executable_utils

FLAGS = flags.FLAGS

flags.DEFINE_string('gtest_executable', None,
                    'The path to the gtest executable')
flags.DEFINE_enum(
    'output_type', 'json', ['json', 'default'],
    """The output type. JSON produces JSON style output with
                  the tests being in an array with the key tests. Default
                  prints all tests separated by line breaks""")
flags.DEFINE_string(
    'output_file', None, """The output path. If not set, all output will get
                    dumped to the terminal""")

flags.mark_flag_as_required('gtest_executable')


def main(_):
  test_list_raw_output = gtest_executable_utils.get_gtest_testlist_raw(
      FLAGS.gtest_executable)
  test_list = gtest_executable_utils.parse_gtest_tests(test_list_raw_output)

  output = ''
  if FLAGS.output_type == 'json':
    test_json = {
        'executable': os.path.basename(FLAGS.gtest_executable),
        'tests': test_list
    }
    output = json.dumps(test_json, indent=4)
  elif FLAGS.output_type == 'default':
    for test in test_list:
      output = output + f'{test}\n'
    # get rid of extra last newline
    output = output[:-1]
  else:
    logging.fatal('output_type should only be json or default')

  if FLAGS.output_file:
    with open(FLAGS.output_file, 'w', encoding='UTF-8') as output_file:
      output_file.write(output)
      print(f'wrote tests to {FLAGS.output_file}', file=sys.stderr)
  else:
    print(output)


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Local ES trainer."""

from absl import app, flags, logging
import gin

from compiler_opt.es import es_trainer_lib

_GIN_FILES = flags.DEFINE_multi_string(
    "gin_files", [], "List of paths to gin configuration files.")
_GIN_BINDINGS = flags.DEFINE_multi_string(
    "gin_bindings", [],
    "Gin bindings to override the values set in the config files.")


def main(_):
  gin.parse_config_files_and_bindings(
      _GIN_FILES.value, bindings=_GIN_BINDINGS.value, skip_unknown=False)
  logging.info(gin.config_str())

  final_weights = es_trainer_lib.train()

  logging.info("Final Weights:")
  logging.info(", ".join(final_weights))


if __name__ == "__main__":
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""util function to save the policy and model config file."""

import dataclasses
import json
import os

from compiler_opt import type_map

import tensorflow as tf
from tf_agents.policies import tf_policy
from tf_agents.policies import policy_saver
from tf_agents.typing import types as tf_agents_types

from typing import Dict, Tuple

OUTPUT_SIGNATURE = 'output_spec.json'
TFLITE_MODEL_NAME = 'model.tflite'

_dtype_to_name_map = {
    dtype: name for name, _, dtype in type_map.TYPE_ASSOCIATIONS
}


def _split_tensor_name(name: str) -> Tuple[str, int]:
  """Return tuple (op, port) with the op and int port for the tensor name."""
  op_port = name.split(':', 2)
  if len(op_port) == 1:
    return op_port[0], 0
  else:
    return op_port[0], int(op_port[1])


# TODO(b/156295309): more sophisticated way of finding tensor names.
def _get_non_identity_op(tensor):
  """Get the true output op aliased by Identity `tensor`.

  Output signature tensors are in a Function that refrences the true call
  in the base SavedModel metagraph.  Traverse the function upwards until
  we find this true output op and tensor and return that.

  Args:
    tensor: A tensor from the unstructured output list of a signature.

  Returns:
    The true associated output tensor of the original function in the main
    SavedModel graph.
  """
  while tensor.op.name.startswith('Identity'):
    tensor = tensor.op.inputs[0]
  return tensor


def convert_saved_model(sm_dir: str, tflite_model_path: str):
  """Convert a saved model to tflite.

  Args:
    sm_dir: path to the saved model to convert

    tflite_model_path: desired output file path. Directory structure will
      be created by this function, as needed.
  """
  tf.io.gfile.makedirs(os.path.dirname(tflite_model_path))
  converter = tf.lite.TFLiteConverter.from_saved_model(sm_dir)
  converter.target_spec.supported_ops = [
      tf.lite.OpsSet.TFLITE_BUILTINS,
  ]
  converter.allow_custom_ops = True
  tfl_model = converter.convert()
  with tf.io.gfile.GFile(tflite_model_path, 'wb') as f:
    f.write(tfl_model)


def convert_mlgo_model(mlgo_model_dir: str, tflite_model_dir: str):
  """Convert a mlgo saved model to mlgo tflite.

  Args:
    mlgo_model_dir: path to the mlgo saved model dir. It is expected to contain
      the saved model files (i.e. saved_model.pb, the variables dir) and the
      output_spec.json file

    tflite_model_dir: path to a directory where the tflite model will be placed.
      The model will be named model.tflite. Alongside it will be placed a copy
      of the output_spec.json file.
  """
  tf.io.gfile.makedirs(tflite_model_dir)
  convert_saved_model(mlgo_model_dir,
                      os.path.join(tflite_model_dir, TFLITE_MODEL_NAME))

  src_json = os.path.join(mlgo_model_dir, OUTPUT_SIGNATURE)
  dest_json = os.path.join(tflite_model_dir, OUTPUT_SIGNATURE)
  tf.io.gfile.copy(src_json, dest_json)


@dataclasses.dataclass(frozen=True)
class Policy:
  """Serialized mlgo policy, used to pass a policy to workers.

  A policy has 2 components, both being file contents:
    - the content of the output_spec.json file;
    - the content of the tflite policy.

  To construct from a directory accessible by tf.io.gfile:

  policy = Policy.from_filesystem(that_dir)

  To make available to the compiler in a directory:

  policy.to_filesystem(that_dir)
  """

  output_spec: bytes
  policy: bytes

  def to_filesystem(self, location: str):
    os.makedirs(location, exist_ok=True)
    output_sig = os.path.join(location, OUTPUT_SIGNATURE)
    policy_path = os.path.join(location, TFLITE_MODEL_NAME)
    with tf.io.gfile.GFile(output_sig, mode='wb') as f:
      f.write(self.output_spec)
    with tf.io.gfile.GFile(policy_path, mode='wb') as f:
      f.write(self.policy)

  @staticmethod
  def from_filesystem(location: str):
    output_sig = os.path.join(location, OUTPUT_SIGNATURE)
    policy_path = os.path.join(location, TFLITE_MODEL_NAME)
    with tf.io.gfile.GFile(output_sig, mode='rb') as f:
      output_spec = f.read()
    with tf.io.gfile.GFile(policy_path, mode='rb') as f:
      policy = f.read()
    return Policy(output_spec=output_spec, policy=policy)


class PolicySaver(object):
  """Object that saves policy and model config file required by inference.

  ```python
  policy_saver = PolicySaver(policy_dict, config)
  policy_saver.save(root_dir)
  ```
  """

  def __init__(self, policy_dict: Dict[str, tf_policy.TFPolicy]):
    """Initialize the PolicySaver object.

    Args:
      policy_dict: A dict mapping from policy name to policy.
    """
    self._policy_saver_dict: Dict[str, Tuple[
        policy_saver.PolicySaver, tf_policy.TFPolicy]] = {
            policy_name: (policy_saver.PolicySaver(
                policy, batch_size=1, use_nest_path_signatures=False), policy)
            for policy_name, policy in policy_dict.items()
        }

  def _write_output_signature(
      self, action_signature: tf_agents_types.NestedTensorSpec, path: str):
    """Writes the output_signature json file into the SavedModel directory."""

    # We'll load the actual SavedModel to be able to map signature names to
    # actual tensor names.
    saved_model = tf.saved_model.load(path)

    # Dict mapping spec name to spec in flattened action signature.
    sm_action_signature = (
        tf.nest.flatten(saved_model.signatures['action'].structured_outputs))

    # Map spec name to index in flattened outputs.
    sm_action_indices = dict(
        (k.name.lower(), i) for i, k in enumerate(sm_action_signature))

    # List mapping flattened structured outputs to tensors.
    sm_action_tensors = saved_model.signatures['action'].outputs

    # First entry in output list is the decision (action)
    decision_spec = tf.nest.flatten(action_signature.action)
    if len(decision_spec) != 1:
      raise ValueError(('Expected action decision to have 1 tensor, but '
                        f'saw: {action_signature.action}'))

    # Find the decision's tensor in the flattened output tensor list.
    sm_action_decision = (
        sm_action_tensors[sm_action_indices[decision_spec[0].name.lower()]])

    sm_action_decision = _get_non_identity_op(sm_action_decision)

    # The first entry in the output_signature file corresponds to the decision.
    (tensor_op, tensor_port) = _split_tensor_name(sm_action_decision.name)
    output_list = [{
        'logging_name': decision_spec[0].name,  # used in SequenceExample.
        'tensor_spec': {
            'name': tensor_op,
            'port': tensor_port,
            'type': _dtype_to_name_map[sm_action_decision.dtype],
            'shape': sm_action_decision.shape.as_list(),
        }
    }]
    for info_spec in tf.nest.flatten(action_signature.info):
      sm_action_info = sm_action_tensors[sm_action_indices[
          info_spec.name.lower()]]
      sm_action_info = _get_non_identity_op(sm_action_info)
      (tensor_op, tensor_port) = _split_tensor_name(sm_action_info.name)
      output_list.append({
          'logging_name': info_spec.name,  # used in SequenceExample.
          'tensor_spec': {
              'name': tensor_op,
              'port': tensor_port,
              'type': _dtype_to_name_map[sm_action_info.dtype],
              'shape': sm_action_info.shape.as_list(),
          }
      })

    with tf.io.gfile.GFile(os.path.join(path, OUTPUT_SIGNATURE), 'w') as f:
      f.write(json.dumps(output_list))

  def save(self, root_dir: str):
    """Writes policy and model_binding.txt to root_dir/policy_name/."""
    for policy_name, (saver, _) in self._policy_saver_dict.items():
      saved_model_dir = os.path.join(root_dir, policy_name)
      saver.save(saved_model_dir)
      self._write_output_signature(saver.policy_step_spec, saved_model_dir)
      # This is not quite the most efficient way to do this - we save the model
      # just to load it again and save it as tflite - but it's the minimum,
      # temporary step so we can validate more thoroughly our use of tflite.
      convert_saved_model(saved_model_dir,
                          os.path.join(saved_model_dir, TFLITE_MODEL_NAME))



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library functions for IR extraction."""

import os
import pathlib
import re
import shutil
import subprocess
import multiprocessing
import functools
import json

from typing import Dict, List, Optional

from absl import logging

from compiler_opt.rl import constant


# TODO(ml-compiler-opt): maybe we can also convert here the cmdline file,from a
# \0 - separated list of strings, to a \n one.
def should_include_module(cmdline: str, match_regexp: Optional[str]) -> bool:
  """Determine if the module should be included."""
  if match_regexp is None:
    return True
  lines = cmdline.split('\0')
  return any(len(re.findall(match_regexp, l)) for l in lines)


def get_thinlto_index(cmdline: str, basedir: str) -> Optional[str]:
  opts = cmdline.split('\0')
  for option in opts:
    if option.startswith('-fthinlto-index'):
      return os.path.join(basedir, option.split('=')[1])
  return None


class TrainingIRExtractor:
  """IR and command line extraction from an object file."""

  def __init__(self, obj_relative_path, output_base_dir, obj_base_dir=None):
    """Set up a TrainingIRExtractor.

    Args:
      obj_relative_path: relative path to the input object file. It will be also
        used to construct the absolute path of the output IR and cmd files, by
        appending it to output_base_dir.
      output_base_dir: the directory under which the output will be produced.
      obj_base_dir: the base directory for all the input object files.
    """
    self._obj_relative_path = obj_relative_path
    self._output_base_dir = output_base_dir
    self._obj_base_dir = obj_base_dir if obj_base_dir is not None else ''

  def obj_base_dir(self):
    return self._obj_base_dir

  def output_base_dir(self):
    return self._output_base_dir

  def relative_output_path(self):
    return self._obj_relative_path

  def input_obj(self):
    return os.path.join(self.obj_base_dir(), self._obj_relative_path)

  def lld_src_bc(self):
    # .3.import.bc is the suffix attached to post-merge-pre-opt ('postimport')
    # IR bitcode saved by lld. It is hardcoded into lld.
    return os.path.join(self._obj_base_dir,
                        self._obj_relative_path + '.3.import.bc')

  def lld_src_thinlto(self):
    return os.path.join(self._obj_base_dir,
                        self._obj_relative_path + '.thinlto.bc')

  def dest_dir(self):
    return os.path.join(self.output_base_dir(),
                        os.path.dirname(self._obj_relative_path))

  def module_name(self):
    return os.path.basename(self._obj_relative_path)

  def cmd_file(self):
    return os.path.join(self.dest_dir(), self.module_name() + '.cmd')

  def bc_file(self):
    return os.path.join(self.dest_dir(), self.module_name() + '.bc')

  def thinlto_index_file(self):
    return os.path.join(self.dest_dir(), self.module_name() + '.thinlto.bc')

  def _get_extraction_cmd_command(self, llvm_objcopy_path: str,
                                  cmd_section_name: str):
    """Get llvm-objcopy and process args to a produce a command string that,
    when invoked, will extract the cmd section info ths self.cmd_file() file.
    """
    return [
        llvm_objcopy_path,
        '--dump-section=' + cmd_section_name + '=' + self.cmd_file(),
        self.input_obj(), '/dev/null'
    ]

  def _get_extraction_bc_command(self, llvm_objcopy_path: str,
                                 bitcode_section_name: str):
    """Gets llvm-objcopy and process args to produce a command string that,
    when invoked, will extract the bitcode section into the self.bc_file()
    file.
    """
    return [
        llvm_objcopy_path,
        '--dump-section=' + bitcode_section_name + '=' + self.bc_file(),
        self.input_obj(), '/dev/null'
    ]

  def _extract_clang_artifacts(self, llvm_objcopy_path: str, cmd_filter: str,
                               is_thinlto: bool, cmd_section_name: str,
                               bitcode_section_name: str) -> Optional[str]:
    """Run llvm-objcopy to extract the .bc and command line."""
    if not os.path.exists(self.input_obj()):
      logging.info('%s does not exist.', self.input_obj())
      return None
    os.makedirs(self.dest_dir(), exist_ok=True)
    try:
      subprocess.check_output(
          self._get_extraction_cmd_command(llvm_objcopy_path, cmd_section_name),
          stderr=subprocess.STDOUT,
          encoding='utf-8')
      if cmd_filter is not None or is_thinlto:
        with open(self.cmd_file(), encoding='utf-8') as f:
          lines = f.readlines()
        assert len(lines) == 1
        cmdline = lines[0]
        if not should_include_module(cmdline, cmd_filter):
          logging.info(
              'Excluding module %s because it does not match the filter',
              self.input_obj())
          os.remove(self.cmd_file())
          return None
        if is_thinlto:
          index_file = get_thinlto_index(cmdline, self.obj_base_dir())
          shutil.copy(index_file, self.thinlto_index_file())

      subprocess.check_output(
          self._get_extraction_bc_command(llvm_objcopy_path,
                                          bitcode_section_name),
          stderr=subprocess.STDOUT,
          encoding='utf-8')
    except subprocess.CalledProcessError as e:
      # This may happen if  .o file was build from asm (.S source).
      logging.warning('%s was not processed: %s', self.input_obj(), e)
      logging.info(e.output)
      return None
    assert (os.path.exists(self.cmd_file()) and
            os.path.exists(self.bc_file()) and
            (not is_thinlto or os.path.exists(self.thinlto_index_file())))
    return self.relative_output_path()

  def _extract_lld_artifacts(self) -> Optional[str]:
    """Extract the .bc file with ThinLTO index from an lld ThinLTO invocation.
    """
    if not os.path.exists(self.lld_src_bc()):
      logging.info('%s does not exist.', self.lld_src_bc())
      return None
    if not os.path.exists(self.lld_src_thinlto()):
      logging.info('%s does not exist.', self.lld_src_thinlto())
      return None
    os.makedirs(self.dest_dir(), exist_ok=True)

    # Copy over the files
    shutil.copy(self.lld_src_bc(), self.bc_file())
    shutil.copy(self.lld_src_thinlto(), self.thinlto_index_file())

    assert os.path.exists(self.bc_file())
    assert os.path.exists(self.thinlto_index_file())
    return self._obj_relative_path

  def extract(self,
              llvm_objcopy_path: Optional[str] = None,
              cmd_filter: Optional[str] = None,
              thinlto_build: Optional[str] = None,
              cmd_section_name: Optional[str] = '.llvmcmd',
              bitcode_section_name: Optional[str] = '.llvmbc') -> Optional[str]:
    if thinlto_build == 'local':
      return self._extract_lld_artifacts()
    return self._extract_clang_artifacts(
        llvm_objcopy_path=llvm_objcopy_path,
        cmd_filter=cmd_filter,
        is_thinlto=thinlto_build == 'distributed',
        cmd_section_name=cmd_section_name,
        bitcode_section_name=bitcode_section_name)


def convert_compile_command_to_objectfile(
    command: Dict[str, str], output_dir: str) -> Optional[TrainingIRExtractor]:
  obj_base_dir = command['directory']
  if 'arguments' in command:
    cmd_parts = command['arguments']
  elif 'command' in command:
    cmd_parts = command['command'].split()
  else:
    logging.info('compile_commands element has no command and arguments')
    return None

  try:
    obj_index = cmd_parts.index('-o') + 1
  except ValueError:
    # This could happen if there are non-clang commands in compile_commands.json
    logging.info('Command has no -o option: %s', ' '.join(cmd_parts))
    return None
  obj_rel_path = cmd_parts[obj_index]
  # TODO(mtrofin): is the obj_base_dir correct for thinlto index bc files?
  return TrainingIRExtractor(
      obj_relative_path=obj_rel_path,
      output_base_dir=output_dir,
      obj_base_dir=obj_base_dir)


def load_from_compile_commands(json_array: List[Dict[str, str]],
                               output_dir: str) -> List[TrainingIRExtractor]:
  objs = [
      convert_compile_command_to_objectfile(cmd, output_dir)
      for cmd in json_array
  ]
  # Filter out None, in case there were non-clang commands in the .json
  return [obj for obj in objs if obj is not None]


def load_from_lld_params(params_array: List[str], obj_base_dir: str,
                         output_dir: str) -> List[TrainingIRExtractor]:
  """Create an ObjectFile array based on lld's parameters."""
  # yank out -o and the output. After that, anything not starting with '-', and
  # ending in a '.o', is an object file.
  try:
    minus_o_idx = params_array.index('-o')
    del params_array[minus_o_idx:minus_o_idx + 2]
    just_obj_paths = [
        o for o in params_array if not o.startswith('-') and o.endswith('.o')
    ]
  except ValueError:
    logging.info('This params file does not have an explicit -o option.')
    just_obj_paths = params_array

  def make_obj(obj_file: str) -> TrainingIRExtractor:
    return TrainingIRExtractor(
        obj_relative_path=obj_file,
        output_base_dir=output_dir,
        obj_base_dir=obj_base_dir)

  return [make_obj(obj_file) for obj_file in just_obj_paths]


def load_from_directory(obj_base_dir: str,
                        output_dir: str) -> List[TrainingIRExtractor]:
  """Create an object file array by globbing an entire drectory.

  Args:
    obj_base_dir: The base build directory that all object files will be
      written out as being relative to.
    output_dir: The output directory where extracted .bc and .cmd files should
      be placed.
  """
  paths = [str(p) for p in pathlib.Path(obj_base_dir).glob('**/*.o')]

  def make_spec(obj_file: str):
    return TrainingIRExtractor(
        obj_relative_path=os.path.relpath(obj_file, start=obj_base_dir),
        output_base_dir=output_dir,
        obj_base_dir=obj_base_dir)

  return [make_spec(path) for path in paths]


def load_for_lld_thinlto(obj_base_dir: str,
                         output_dir: str) -> List[TrainingIRExtractor]:
  # .3.import.bc is the suffix attached to post-merge-pre-opt ('postimport')
  # IR bitcode saved by lld. It is hardcoded into lld. ThinLTO index files
  # are also emitted next to the postimport bitcode, with the suffix
  # .thinlto.bc instead
  paths = [str(p) for p in pathlib.Path(obj_base_dir).glob('**/*.3.import.bc')]

  def make_spec(obj_file: str):
    return TrainingIRExtractor(
        # Cut away .3.import.bc
        obj_relative_path=os.path.relpath(obj_file, start=obj_base_dir)[:-12],
        output_base_dir=output_dir,
        obj_base_dir=obj_base_dir)

  return [make_spec(path) for path in paths]


def run_extraction(objs: List[TrainingIRExtractor], num_workers: int,
                   llvm_objcopy_path: str, cmd_filter: str, thinlto_build: str,
                   cmd_section_name: str, bitcode_section_name: str):
  """Extracts all specified object files into the corpus directory.

  Args:
    objs: A list of TrainingIRExtractor Objects that represent the object files
      to extract bitcode/commands from.
    num_workers: The number of parallel processes to spawn to run the
      extraction.
    llvm_objcopy_path: The path to the llvm-objcopy to use for dumping sections.
    cmd_filter: A regular expression that is used to select for compilations
      performed with specific flags. If you want to include all compilations,
      set this to None.
    thinlto_build: Whether or not this is a ThinLTO build, and if so, the type.
      Set this to None if the build was not done with ThinLTO.
    cmd_section_name: The name of the command line section created by the
      bitcode embedding.
    bitcode_section_name: The name of the bitcode section created by the
      bitcode embedding.
  """
  extract_artifacts = functools.partial(
      TrainingIRExtractor.extract,
      llvm_objcopy_path=llvm_objcopy_path,
      cmd_filter=cmd_filter,
      thinlto_build=thinlto_build,
      cmd_section_name=cmd_section_name,
      bitcode_section_name=bitcode_section_name)

  with multiprocessing.Pool(num_workers) as pool:
    relative_output_paths = pool.map(extract_artifacts, objs)
    pool.close()
    pool.join()
  return relative_output_paths


def write_corpus_manifest(thinlto_build: str, relative_output_paths: List[str],
                          output_dir: str):
  """Writes a corpus_manifest.json containing all necessary information about
  the corpus.

  Args:
    thinlto_build: Whether or not the build was done with ThinLTO and if so,
      what kind of ThinLTO. Set this to none if the build was not performed with
      ThinLTO.
    relative_output_paths: The relative (to the corpus directory) output paths
      of all the bitcode files that should be placed in the corpus manifest
    output_dir: The corpus directory where the corpus manifest should be
      placed.
  """
  # This comes first rather than later so global_command_override is at the top
  # of the .json after being written
  if thinlto_build == 'local':
    corpus_description = {
        'global_command_override': constant.UNSPECIFIED_OVERRIDE
    }
  else:
    corpus_description = {}

  corpus_description.update({
      'has_thinlto': thinlto_build is not None,
      'modules': [path for path in relative_output_paths if path is not None]
  })

  with open(
      os.path.join(output_dir, 'corpus_description.json'),
      'w',
      encoding='utf-8') as f:
    json.dump(corpus_description, f, indent=2)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Local ES trainer."""

from absl import flags, logging
import functools
import gin
import tensorflow as tf
import os

from compiler_opt.distributed.local import local_worker_manager
from compiler_opt.es import blackbox_optimizers
from compiler_opt.es import gradient_ascent_optimization_algorithms
from compiler_opt.es import blackbox_learner
from compiler_opt.es import policy_utils
from compiler_opt.rl import policy_saver, corpus

POLICY_NAME = "policy"

FLAGS = flags.FLAGS

_BETA1 = flags.DEFINE_float("beta1", 0.9,
                            "Beta1 for ADAM gradient ascent optimizer.")
_BETA2 = flags.DEFINE_float("beta2", 0.999,
                            "Beta2 for ADAM gradient ascent optimizer.")
_GRAD_REG_ALPHA = flags.DEFINE_float(
    "grad_reg_alpha", 0.01,
    "Weight of regularization term in regression gradient.")
_GRAD_REG_TYPE = flags.DEFINE_string(
    "grad_reg_type", "ridge",
    "Regularization method to use with regression gradient.")
_GRADIENT_ASCENT_OPTIMIZER_TYPE = flags.DEFINE_string(
    "gradient_ascent_optimizer_type", None,
    "Gradient ascent optimization algorithm: 'momentum' or 'adam'")
flags.mark_flag_as_required("gradient_ascent_optimizer_type")
_GREEDY = flags.DEFINE_bool(
    "greedy",
    None,
    "Whether to construct a greedy policy (argmax). \
      If False, a sampling-based policy will be used.",
    required=True)
_MOMENTUM = flags.DEFINE_float(
    "momentum", 0.0, "Momentum for momentum gradient ascent optimizer.")
_OUTPUT_PATH = flags.DEFINE_string("output_path", "",
                                   "Path to write all output")
_PRETRAINED_POLICY_PATH = flags.DEFINE_string(
    "pretrained_policy_path", None,
    "The path of the pretrained policy. If not provided, it will \
        construct a new policy with randomly initialized weights.")
_REQUEST_DEADLINE = flags.DEFINE_float(
    "request_deadline", 30.0, "Deadline in seconds for requests \
    to the data collection requests.")
_TRAIN_CORPORA = flags.DEFINE_string("train_corpora", "",
                                     "List of paths to training corpora")


@gin.configurable
def train(additional_compilation_flags=(),
          delete_compilation_flags=(),
          worker_class=None):
  """Train with ES."""

  if not _TRAIN_CORPORA.value:
    raise ValueError("Need to supply nonempty train corpora.")

  # Create directories
  if not tf.io.gfile.isdir(_OUTPUT_PATH.value):
    tf.io.gfile.makedirs(_OUTPUT_PATH.value)

  # Construct the policy and upload it
  policy = policy_utils.create_actor_policy(greedy=_GREEDY.value)
  saver = policy_saver.PolicySaver({POLICY_NAME: policy})

  # Save the policy
  policy_save_path = os.path.join(_OUTPUT_PATH.value, "policy")
  saver.save(policy_save_path)

  # Get initial parameter
  if not _PRETRAINED_POLICY_PATH.value:
    # Use randomly initialized parameters
    logging.info("Use random parameters")
    initial_parameters = policy_utils.get_vectorized_parameters_from_policy(
        policy)
    logging.info("Parameter dimension: %s", initial_parameters.shape)
    logging.info("Initial parameters: %s", initial_parameters)
  else:
    # Read the parameters from the pretrained policy
    logging.info("Reading policy parameters from %s",
                 _PRETRAINED_POLICY_PATH.value)
    # Load the policy
    pretrained_policy = tf.saved_model.load(_PRETRAINED_POLICY_PATH.value)
    initial_parameters = policy_utils.get_vectorized_parameters_from_policy(
        pretrained_policy)

  policy_parameter_dimension = (
      policy_utils.get_vectorized_parameters_from_policy(policy).shape[0])
  if policy_parameter_dimension != initial_parameters.shape[0]:
    raise ValueError("Pretrained policy dimension is incorrect")

  logging.info("Parameter dimension: %s", initial_parameters.shape)
  logging.info("Initial parameters: %s", initial_parameters)

  cps = corpus.create_corpus_for_testing(
      location=_TRAIN_CORPORA.value,
      elements=[corpus.ModuleSpec(name="smth", size=1)],
      additional_flags=additional_compilation_flags,
      delete_flags=delete_compilation_flags)

  # Construct policy saver
  saved_policy = policy_utils.create_actor_policy(greedy=True)
  policy_saver_function = functools.partial(
      policy_utils.save_policy,
      policy=saved_policy,
      save_folder=os.path.join(_OUTPUT_PATH.value, "saved_policies"))

  # Get learner config
  learner_config = blackbox_learner.BlackboxLearnerConfig()

  # the following are from Blackbox Library.
  init_current_input = initial_parameters
  init_iteration = 0
  metaparams = []  # Ignore meta params for state normalization for now
  # TODO(linzinan): delete all unused parameters.

  # ------------------ GRADIENT ASCENT OPTIMIZERS ------------------------------
  if _GRADIENT_ASCENT_OPTIMIZER_TYPE.value == "momentum":
    logging.info("Running momentum gradient ascent optimizer")
    # You can obtain a vanilla gradient ascent optimizer by setting momentum=0.0
    # and setting step_size to the desired learning rate.
    gradient_ascent_optimizer = (
        gradient_ascent_optimization_algorithms.MomentumOptimizer(
            learner_config.step_size, _MOMENTUM.value))
  elif _GRADIENT_ASCENT_OPTIMIZER_TYPE.value == "adam":
    logging.info("Running Adam gradient ascent optimizer")
    gradient_ascent_optimizer = (
        gradient_ascent_optimization_algorithms.AdamOptimizer(
            learner_config.step_size, _BETA1.value, _BETA2.value))
  else:
    logging.info("No gradient ascent \
                 optimizer selected. Stopping.")
    return
  # ----------------------------------------------------------------------------

  # ------------------ OPTIMIZERS ----------------------------------------------
  if learner_config.blackbox_optimizer == (
      blackbox_optimizers.Algorithm.MONTE_CARLO):
    logging.info("Running ES/ARS. Filtering: %s directions",
                 str(learner_config.num_top_directions))
    blackbox_optimizer = blackbox_optimizers.MonteCarloBlackboxOptimizer(
        learner_config.precision_parameter, learner_config.est_type,
        learner_config.fvalues_normalization,
        learner_config.hyperparameters_update_method, metaparams, None,
        learner_config.num_top_directions, gradient_ascent_optimizer)
  elif learner_config.blackbox_optimizer == (
      blackbox_optimizers.Algorithm.TRUST_REGION):
    logging.info("Running trust region")
    tr_params = {
        "init_radius": FLAGS.tr_init_radius,
        "grow_threshold": FLAGS.tr_grow_threshold,
        "grow_factor": FLAGS.tr_grow_factor,
        "shrink_neg_threshold": FLAGS.tr_shrink_neg_threshold,
        "shrink_factor": FLAGS.tr_shrink_factor,
        "reject_threshold": FLAGS.tr_reject_threshold,
        "reject_factor": FLAGS.tr_reject_factor,
        "dense_hessian": FLAGS.tr_dense_hessian,
        "sub_termination": FLAGS.tr_sub_termination,
        "subproblem_maxiter": FLAGS.tr_subproblem_maxiter,
        "minimum_radius": FLAGS.tr_minimum_radius,
        "grad_type": FLAGS.grad_type,
        "grad_reg_type": _GRAD_REG_TYPE.value,
        "grad_reg_alpha": _GRAD_REG_ALPHA.value
    }
    for param, value in tr_params.items():
      logging.info("%s: %s", param, value)
      blackbox_optimizer = blackbox_optimizers.TrustRegionOptimizer(
          learner_config.precision_parameter, learner_config.est_type,
          learner_config.fvalues_normalization,
          learner_config.hyperparameters_update_method, metaparams, tr_params)
  elif learner_config.blackbox_optimizer == (
      blackbox_optimizers.Algorithm.SKLEARN_REGRESSION):
    logging.info("Running Regression Based Optimizer")
    blackbox_optimizer = blackbox_optimizers.SklearnRegressionBlackboxOptimizer(
        _GRAD_REG_TYPE.value, _GRAD_REG_ALPHA.value, learner_config.est_type,
        learner_config.fvalues_normalization,
        learner_config.hyperparameters_update_method, metaparams, None,
        gradient_ascent_optimizer)
  else:
    raise ValueError(
        f"Unknown optimizer: '{learner_config.blackbox_optimizer}'")

  logging.info("Initializing blackbox learner.")
  learner = blackbox_learner.BlackboxLearner(
      blackbox_opt=blackbox_optimizer,
      sampler=cps,
      tf_policy_path=os.path.join(policy_save_path, POLICY_NAME),
      output_dir=_OUTPUT_PATH.value,
      policy_saver_fn=policy_saver_function,
      model_weights=init_current_input,
      config=learner_config,
      initial_step=init_iteration,
      deadline=_REQUEST_DEADLINE.value)

  if not worker_class:
    logging.info("No Worker class selected. Stopping.")
    return

  logging.info("Ready to train: running for %d steps.",
               learner_config.total_steps)

  with local_worker_manager.LocalWorkerPoolManager(
      worker_class, learner_config.total_num_perturbations, arg="",
      kwarg="") as pool:
    for _ in range(learner_config.total_steps):
      learner.run_step(pool)

  return learner.get_model_weights()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compiler_opt.rl.local_data_collector."""

# pylint: disable=protected-access
import collections
import string
import sys
from typing import List, Tuple

import tensorflow as tf
from tf_agents.system import system_multiprocessing as multiprocessing

# This is https://github.com/google/pytype/issues/764
from google.protobuf import text_format  # pytype: disable=pyi-error
from compiler_opt.distributed.local.local_worker_manager import LocalWorkerPoolManager
from compiler_opt.rl import compilation_runner
from compiler_opt.rl import corpus
from compiler_opt.rl import data_collector
from compiler_opt.rl import local_data_collector
from compiler_opt.rl import policy_saver

_policy_str = 'policy'.encode(encoding='utf-8')

_mock_policy = policy_saver.Policy(output_spec=bytes(), policy=_policy_str)


def _get_sequence_example(feature_value):
  sequence_example_text = string.Template("""
    feature_lists {
      feature_list {
        key: "feature_0"
        value {
          feature { int64_list { value: $feature_value } }
          feature { int64_list { value: $feature_value } }
        }
      }
    }""").substitute(feature_value=feature_value)
  return text_format.Parse(sequence_example_text, tf.train.SequenceExample())


def mock_collect_data(loaded_module_spec: corpus.LoadedModuleSpec, policy,
                      reward_stat, model_id):
  assert loaded_module_spec.name.startswith('dummy')
  assert policy.policy == _policy_str
  assert reward_stat is None or reward_stat == {
      'default':
          compilation_runner.RewardStat(
              default_reward=1, moving_average_reward=2)
  }
  if reward_stat is None:
    return compilation_runner.CompilationResult(
        sequence_examples=[_get_sequence_example(feature_value=1)],
        reward_stats={
            'default':
                compilation_runner.RewardStat(
                    default_reward=1, moving_average_reward=2)
        },
        rewards=[1.2],
        policy_rewards=[36],
        keys=['default'],
        model_id=model_id)
  else:
    return compilation_runner.CompilationResult(
        sequence_examples=[_get_sequence_example(feature_value=2)],
        reward_stats={
            'default':
                compilation_runner.RewardStat(
                    default_reward=1, moving_average_reward=3)
        },
        rewards=[3.4],
        policy_rewards=[18],
        keys=['default'],
        model_id=model_id)


class Sleeper(compilation_runner.CompilationRunner):
  """Test CompilationRunner that just sleeps."""

  def collect_data(self,
                   loaded_module_spec,
                   policy=None,
                   reward_stat=None,
                   model_id=None):
    _ = loaded_module_spec, policy, reward_stat
    compilation_runner.start_cancellable_process(['sleep', '3600s'], 3600,
                                                 self._cancellation_manager)

    return compilation_runner.CompilationResult(
        sequence_examples=[],
        reward_stats={},
        rewards=[],
        policy_rewards=[],
        keys=[],
        model_id=model_id)


class MyRunner(compilation_runner.CompilationRunner):

  def collect_data(self, *args, **kwargs):
    return mock_collect_data(*args, **kwargs)


class DeterministicSampler(corpus.Sampler):
  """A corpus sampler that returns modules in order, and can also be reset."""

  def __init__(self, module_specs: Tuple[corpus.ModuleSpec]):
    super().__init__(module_specs)
    self._cur_pos = 0

  def __call__(self, k: int, n: int = 20) -> List[corpus.ModuleSpec]:
    ret = []
    for _ in range(k):
      ret.append(self._module_specs[self._cur_pos % len(self._module_specs)])
      self._cur_pos += 1
    return ret

  def reset(self):
    self._cur_pos = 0


class LocalDataCollectorTest(tf.test.TestCase):

  def test_local_data_collector(self):

    def create_test_iterator_fn():

      def _test_iterator_fn(data_list):
        assert data_list in (
            [_get_sequence_example(feature_value=1).SerializeToString()] * 9,
            [_get_sequence_example(feature_value=2).SerializeToString()] * 9)
        if data_list == [
            _get_sequence_example(feature_value=1).SerializeToString()
        ] * 9:
          return iter(tf.data.Dataset.from_tensor_slices([1, 2, 3]))
        else:
          return iter(tf.data.Dataset.from_tensor_slices([4, 5, 6]))

      return _test_iterator_fn

    with LocalWorkerPoolManager(worker_class=MyRunner, count=4) as lwp:
      cps = corpus.create_corpus_for_testing(
          location=self.create_tempdir(),
          elements=[
              corpus.ModuleSpec(name=f'dummy{i}', size=i) for i in range(100)
          ],
          sampler_type=DeterministicSampler)
      collector = local_data_collector.LocalDataCollector(
          cps=cps,
          num_modules=9,
          worker_pool=lwp,
          parser=create_test_iterator_fn(),
          reward_stat_map=collections.defaultdict(lambda: None),
          best_trajectory_repo=None)

      # reset the sampler, so the next time we collect, we collect the same
      # modules. We do it before the collect_data call, because that's when
      # we'll re-sample to prefetch the next batch.
      cps.reset()

      data_iterator, monitor_dict = collector.collect_data(
          policy=_mock_policy, model_id=0)
      data = list(data_iterator)
      self.assertEqual([1, 2, 3], data)
      expected_monitor_dict_subset = {
          'default': {
              'success_modules': 9,
              'total_trajectory_length': 18,
          }
      }
      # Issue #38
      if sys.version_info.minor >= 9:
        self.assertEqual(monitor_dict,
                         monitor_dict | expected_monitor_dict_subset)
      else:
        self.assertEqual(monitor_dict, {
            **monitor_dict,
            **expected_monitor_dict_subset
        })
      data_iterator, monitor_dict = collector.collect_data(
          policy=_mock_policy, model_id=0)
      data = list(data_iterator)
      # because we reset the sampler, these are the same modules
      self.assertEqual([4, 5, 6], data)
      expected_monitor_dict_subset = {
          'default': {
              'success_modules': 9,
              'total_trajectory_length': 18,
          }
      }
      # Issue #38
      if sys.version_info.minor >= 9:
        self.assertEqual(monitor_dict,
                         monitor_dict | expected_monitor_dict_subset)
      else:
        self.assertEqual(monitor_dict, {
            **monitor_dict,
            **expected_monitor_dict_subset
        })

      collector.close_pool()

  def test_local_data_collector_task_management(self):

    def parser(_):
      pass

    class QuickExiter(data_collector.EarlyExitChecker):

      def __init__(self, num_modules):
        data_collector.EarlyExitChecker.__init__(self, num_modules=num_modules)

      def wait(self, _):
        return False

    with LocalWorkerPoolManager(worker_class=Sleeper, count=4) as lwp:
      collector = local_data_collector.LocalDataCollector(
          cps=corpus.create_corpus_for_testing(
              location=self.create_tempdir(),
              elements=[
                  corpus.ModuleSpec(name=f'dummy{i}', size=1)
                  for i in range(200)
              ]),
          num_modules=4,
          worker_pool=lwp,
          parser=parser,
          reward_stat_map=collections.defaultdict(lambda: None),
          best_trajectory_repo=None,
          exit_checker_ctor=QuickExiter)
      collector.collect_data(policy=_mock_policy, model_id=0)
      collector._join_pending_jobs()
      killed = 0
      for w in collector._current_futures:
        self.assertRaises(compilation_runner.ProcessKilledError, w.result)
        killed += 1
      self.assertEqual(killed, 4)
      collector.close_pool()


if __name__ == '__main__':
  multiprocessing.handle_test_main(tf.test.main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compiler_opt.rl.policy_saver."""

import filecmp
import json
import os

import tensorflow as tf

from tf_agents.agents.behavioral_cloning import behavioral_cloning_agent
from tf_agents.networks import q_rnn_network
from tf_agents.specs import tensor_spec
from tf_agents.trajectories import time_step

from compiler_opt.rl import policy_saver


# copied from the llvm regalloc generator
def _gen_test_model(outdir: str):
  policy_decision_label = 'index_to_evict'
  policy_output_spec = """
  [
      {
          "logging_name": "index_to_evict",
          "tensor_spec": {
              "name": "StatefulPartitionedCall",
              "port": 0,
              "type": "int64_t",
              "shape": [
                  1
              ]
          }
      }
  ]
  """
  per_register_feature_list = ['mask']
  num_registers = 33

  def get_input_signature():
    """Returns (time_step_spec, action_spec) for LLVM register allocation."""
    inputs = dict(
        (key, tf.TensorSpec(dtype=tf.int64, shape=(num_registers), name=key))
        for key in per_register_feature_list)
    return inputs

  module = tf.Module()
  # We have to set this useless variable in order for the TF C API to correctly
  # intake it
  module.var = tf.Variable(0, dtype=tf.int64)

  def action(*inputs):
    result = tf.math.argmax(
        tf.cast(inputs[0]['mask'], tf.int32), axis=-1) + module.var
    return {policy_decision_label: result}

  module.action = tf.function()(action)
  action = {
      'action': module.action.get_concrete_function(get_input_signature())
  }
  tf.saved_model.save(module, outdir, signatures=action)
  output_spec_path = os.path.join(outdir, 'output_spec.json')
  with tf.io.gfile.GFile(output_spec_path, 'w') as f:
    print(f'Writing output spec to {output_spec_path}.')
    f.write(policy_output_spec)


class PolicySaverTest(tf.test.TestCase):

  def setUp(self):
    super().setUp()
    observation_spec = tf.TensorSpec(
        dtype=tf.int64, shape=(), name='callee_users')
    self._time_step_spec = time_step.time_step_spec(observation_spec)
    self._action_spec = tensor_spec.BoundedTensorSpec(
        dtype=tf.int64,
        shape=(),
        minimum=0,
        maximum=1,
        name='inlining_decision')
    self._network = q_rnn_network.QRnnNetwork(
        input_tensor_spec=self._time_step_spec.observation,
        action_spec=self._action_spec,
        lstm_size=(40,))

  def test_save_policy(self):
    test_agent = behavioral_cloning_agent.BehavioralCloningAgent(
        self._time_step_spec, self._action_spec, self._network,
        tf.compat.v1.train.AdamOptimizer())
    policy_dict = {
        'saved_policy': test_agent.policy,
        'saved_collect_policy': test_agent.collect_policy
    }
    test_policy_saver = policy_saver.PolicySaver(policy_dict=policy_dict)

    root_dir = self.get_temp_dir()
    test_policy_saver.save(root_dir)

    sub_dirs = tf.io.gfile.listdir(root_dir)
    self.assertCountEqual(['saved_policy', 'saved_collect_policy'], sub_dirs)

    for sub_dir in ['saved_policy', 'saved_collect_policy']:
      self.assertTrue(
          tf.io.gfile.exists(os.path.join(root_dir, sub_dir, 'saved_model.pb')))
      self.assertTrue(
          tf.io.gfile.exists(os.path.join(root_dir, sub_dir, 'model.tflite')))
      self.assertTrue(
          tf.io.gfile.exists(
              os.path.join(root_dir, sub_dir,
                           'variables/variables.data-00000-of-00001')))
      output_signature_fn = os.path.join(root_dir, sub_dir, 'output_spec.json')
      self.assertTrue(tf.io.gfile.exists(output_signature_fn))
      self.assertEqual([{
          'logging_name': 'inlining_decision',
          'tensor_spec': {
              'name': 'StatefulPartitionedCall',
              'port': 0,
              'type': 'int64_t',
              'shape': [1],
          }
      }], json.loads(tf.io.gfile.GFile(output_signature_fn).read()))

  def test_tflite_conversion(self):
    sm_dir = os.path.join(self.get_temp_dir(), 'saved_model')
    tflite_dir = os.path.join(self.get_temp_dir(), 'tflite_model')
    _gen_test_model(sm_dir)
    policy_saver.convert_mlgo_model(sm_dir, tflite_dir)
    self.assertTrue(
        tf.io.gfile.exists(
            os.path.join(tflite_dir, policy_saver.TFLITE_MODEL_NAME)))
    self.assertTrue(
        tf.io.gfile.exists(
            os.path.join(tflite_dir, policy_saver.OUTPUT_SIGNATURE)))

  def test_policy_serialization(self):
    sm_dir = os.path.join(self.get_temp_dir(), 'model')
    orig_dir = os.path.join(self.get_temp_dir(), 'orig_model')
    dest_dir = os.path.join(self.get_temp_dir(), 'dest_model')
    _gen_test_model(sm_dir)
    policy_saver.convert_mlgo_model(sm_dir, orig_dir)

    serialized_policy = policy_saver.Policy.from_filesystem(orig_dir)
    serialized_policy.to_filesystem(dest_dir)

    self.assertTrue(
        filecmp.cmp(
            os.path.join(orig_dir, policy_saver.TFLITE_MODEL_NAME),
            os.path.join(dest_dir, policy_saver.TFLITE_MODEL_NAME),
            shallow=False))
    self.assertTrue(
        filecmp.cmp(
            os.path.join(orig_dir, policy_saver.OUTPUT_SIGNATURE),
            os.path.join(dest_dir, policy_saver.OUTPUT_SIGNATURE),
            shallow=False))


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compiler_opt.rl.trainer."""

# pylint: disable=protected-access

import tensorflow as tf
from tf_agents.agents.behavioral_cloning import behavioral_cloning_agent
from tf_agents.networks import q_rnn_network
from tf_agents.specs import tensor_spec
from tf_agents.trajectories import time_step
from tf_agents.trajectories import trajectory
from unittest import mock

from compiler_opt.rl import trainer


def _create_test_data(batch_size, sequence_length):
  test_trajectory = trajectory.Trajectory(
      step_type=tf.fill([batch_size, sequence_length], 1),
      observation={
          'inlining_default':
              tf.fill([batch_size, sequence_length],
                      tf.constant(10, dtype=tf.int64))
      },
      action=tf.fill([batch_size, sequence_length],
                     tf.constant(1, dtype=tf.int64)),
      policy_info=(),
      next_step_type=tf.fill([batch_size, sequence_length], 1),
      reward=tf.fill([batch_size, sequence_length], 2.0),
      discount=tf.fill([batch_size, sequence_length], 1.0),
  )

  def test_data_iterator():
    while True:
      yield test_trajectory

  return test_data_iterator()


class TrainerTest(tf.test.TestCase):

  def setUp(self):
    observation_spec = {
        'inlining_default':
            tf.TensorSpec(dtype=tf.int64, shape=(), name='inlining_default')
    }
    self._time_step_spec = time_step.time_step_spec(observation_spec)
    self._action_spec = tensor_spec.BoundedTensorSpec(
        dtype=tf.int64,
        shape=(),
        minimum=0,
        maximum=1,
        name='inlining_decision')
    self._network = q_rnn_network.QRnnNetwork(
        input_tensor_spec=self._time_step_spec.observation,
        action_spec=self._action_spec,
        lstm_size=(40,),
        preprocessing_layers={
            'inlining_default': tf.keras.layers.Lambda(lambda x: x)
        })
    super().setUp()

  def test_trainer_initialization(self):
    test_agent = behavioral_cloning_agent.BehavioralCloningAgent(
        self._time_step_spec,
        self._action_spec,
        self._network,
        tf.compat.v1.train.AdamOptimizer(),
        num_outer_dims=2)
    test_trainer = trainer.Trainer(
        root_dir=self.get_temp_dir(), agent=test_agent)
    self.assertEqual(0, test_trainer._global_step.numpy())

  def test_training(self):
    test_agent = behavioral_cloning_agent.BehavioralCloningAgent(
        self._time_step_spec,
        self._action_spec,
        self._network,
        tf.compat.v1.train.AdamOptimizer(),
        num_outer_dims=2)
    test_trainer = trainer.Trainer(
        root_dir=self.get_temp_dir(),
        agent=test_agent,
        summary_log_interval=1,
        summary_export_interval=10)
    self.assertEqual(0, test_trainer._global_step.numpy())

    dataset_iter = _create_test_data(batch_size=3, sequence_length=3)
    monitor_dict = {'default': {'test': 1}}

    with mock.patch.object(
        tf.summary, 'scalar', autospec=True) as mock_scalar_summary:
      test_trainer.train(dataset_iter, monitor_dict, num_iterations=100)
      self.assertEqual(
          10,
          sum(1 for c in mock_scalar_summary.mock_calls
              if c[2]['name'] == 'test'))
      self.assertEqual(100, test_trainer._global_step.numpy())
      self.assertEqual(100, test_trainer.global_step_numpy())

  def test_training_with_multiple_times(self):
    test_agent = behavioral_cloning_agent.BehavioralCloningAgent(
        self._time_step_spec,
        self._action_spec,
        self._network,
        tf.compat.v1.train.AdamOptimizer(),
        num_outer_dims=2)
    test_trainer = trainer.Trainer(
        root_dir=self.get_temp_dir(), agent=test_agent)
    self.assertEqual(0, test_trainer._global_step.numpy())

    dataset_iter = _create_test_data(batch_size=3, sequence_length=3)
    monitor_dict = {'default': {'test': 1}}
    test_trainer.train(dataset_iter, monitor_dict, num_iterations=10)
    self.assertEqual(10, test_trainer._global_step.numpy())

    dataset_iter = _create_test_data(batch_size=6, sequence_length=4)
    test_trainer.train(dataset_iter, monitor_dict, num_iterations=10)
    self.assertEqual(20, test_trainer._global_step.numpy())

  def test_inference(self):
    test_agent = behavioral_cloning_agent.BehavioralCloningAgent(
        self._time_step_spec,
        self._action_spec,
        self._network,
        tf.compat.v1.train.AdamOptimizer(),
        num_outer_dims=2)
    test_trainer = trainer.Trainer(
        root_dir=self.get_temp_dir(), agent=test_agent)

    inference_batch_size = 1
    random_time_step = tensor_spec.sample_spec_nest(
        self._time_step_spec, outer_dims=(inference_batch_size,))

    initial_policy_state = test_trainer._agent.policy.get_initial_state(
        inference_batch_size)

    action_outputs = test_trainer._agent.policy.action(random_time_step,
                                                       initial_policy_state)
    self.assertAllEqual([inference_batch_size], action_outputs.action.shape)

    action_outputs = test_trainer._agent.policy.action(random_time_step,
                                                       action_outputs.state)
    self.assertAllEqual([inference_batch_size], action_outputs.action.shape)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Util functions to create and edit a tf_agent policy."""

from typing import Protocol, Sequence, Type

import gin
import numpy as np
import numpy.typing as npt
import tensorflow as tf
from tf_agents.networks import network
from tf_agents.policies import actor_policy
from tf_agents.policies import greedy_policy
from tf_agents.policies import tf_policy

from compiler_opt.rl import policy_saver
from compiler_opt.rl import registry


class HasModelVariables(Protocol):
  model_variables: Sequence[tf.Variable]


# TODO(abenalaast): Issue #280
@gin.configurable(module='policy_utils')
def create_actor_policy(
    actor_network_ctor: Type[network.DistributionNetwork],
    greedy: bool = False,
) -> tf_policy.TFPolicy:
  """Creates an actor policy."""
  problem_config = registry.get_configuration()
  time_step_spec, action_spec = problem_config.get_signature_spec()
  layers = tf.nest.map_structure(
      problem_config.get_preprocessing_layer_creator(),
      time_step_spec.observation)

  actor_network = actor_network_ctor(
      input_tensor_spec=time_step_spec.observation,
      output_tensor_spec=action_spec,
      preprocessing_layers=layers)

  policy = actor_policy.ActorPolicy(
      time_step_spec=time_step_spec,
      action_spec=action_spec,
      actor_network=actor_network)

  if greedy:
    policy = greedy_policy.GreedyPolicy(policy)

  return policy


def get_vectorized_parameters_from_policy(
    policy: 'tf_policy.TFPolicy | HasModelVariables'
) -> npt.NDArray[np.float32]:
  """Returns a policy's variable values as a single np array."""
  if isinstance(policy, tf_policy.TFPolicy):
    variables = policy.variables()
  elif hasattr(policy, 'model_variables'):
    variables = policy.model_variables
  else:
    raise ValueError(f'Policy must be a TFPolicy or a loaded SavedModel. '
                     f'Passed policy: {policy}')

  parameters = [var.numpy().flatten() for var in variables]
  parameters = np.concatenate(parameters, axis=0)
  return parameters


def set_vectorized_parameters_for_policy(
    policy: 'tf_policy.TFPolicy | HasModelVariables',
    parameters: npt.NDArray[np.float32]) -> None:
  """Separates values in parameters into the policy's shapes
  and sets the policy variables to those values"""
  if isinstance(policy, tf_policy.TFPolicy):
    variables = policy.variables()
  elif hasattr(policy, 'model_variables'):
    variables = policy.model_variables
  else:
    raise ValueError(f'Policy must be a TFPolicy or a loaded SavedModel. '
                     f'Passed policy: {policy}')

  param_pos = 0
  for variable in variables:
    shape = tf.shape(variable).numpy()
    num_elems = np.prod(shape)
    param = np.reshape(parameters[param_pos:param_pos + num_elems], shape)
    variable.assign(param)
    param_pos += num_elems
  if param_pos != len(parameters):
    raise ValueError(
        f'Parameter dimensions are not matched! Expected {len(parameters)} '
        f'but only found {param_pos}.')


def save_policy(policy: 'tf_policy.TFPolicy | HasModelVariables',
                parameters: npt.NDArray[np.float32], save_folder: str,
                policy_name: str) -> None:
  """Assigns a policy the name policy_name
  and saves it to the directory of save_folder
  with the values in parameters."""
  set_vectorized_parameters_for_policy(policy, parameters)
  saver = policy_saver.PolicySaver({policy_name: policy})
  saver.save(save_folder)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Map between tf, ctypes, and string names for scalar types."""
import ctypes
from typing import List, Tuple, Union
import tensorflow as tf

ScalarCType = Union['type[ctypes.c_float]', 'type[ctypes.c_double]',
                    'type[ctypes.c_int8]', 'type[ctypes.c_int16]',
                    'type[ctypes.c_uint16]', 'type[ctypes.c_int32]',
                    'type[ctypes.c_uint32]', 'type[ctypes.c_int64]',
                    'type[ctypes.c_uint64]']

TYPE_ASSOCIATIONS: List[Tuple[str, ScalarCType,
                              tf.DType]] = [
                                  ('float', ctypes.c_float, tf.float32),
                                  ('double', ctypes.c_double, tf.float64),
                                  ('int8_t', ctypes.c_int8, tf.int8),
                                  ('uint8_t', ctypes.c_uint8, tf.uint8),
                                  ('int16_t', ctypes.c_int16, tf.int16),
                                  ('uint16_t', ctypes.c_uint16, tf.uint16),
                                  ('int32_t', ctypes.c_int32, tf.int32),
                                  ('uint32_t', ctypes.c_uint32, tf.uint32),
                                  ('int64_t', ctypes.c_int64, tf.int64),
                                  ('uint64_t', ctypes.c_uint64, tf.uint64)
                              ]



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""operations to transform features (observations)."""

import os
import re

from typing import List, Callable, Optional

import numpy as np
import tensorflow.compat.v2 as tf
from tf_agents.typing import types


def build_quantile_map(quantile_file_dir: str):
  """build feature quantile map by reading from files in quantile_file_dir."""
  quantile_map = {}
  pattern = os.path.join(quantile_file_dir, '(.*).buckets')
  for quantile_file_path in tf.io.gfile.glob(
      os.path.join(quantile_file_dir, '*.buckets')):
    m = re.fullmatch(pattern, quantile_file_path)
    assert m
    feature_name = m.group(1)
    with tf.io.gfile.GFile(quantile_file_path, 'r') as quantile_file:
      raw_quantiles = [float(x) for x in quantile_file]
    quantile_map[feature_name] = raw_quantiles

  return quantile_map


def discard_fn(obs: types.Float):
  """discard the input feature by setting it to 0."""
  zeros = tf.zeros_like(obs, dtype=tf.float32)
  return tf.expand_dims(zeros, axis=-1)


def identity_fn(obs: types.Float):
  """Return the same value with expanding the last dimension."""
  return tf.cast(tf.expand_dims(obs, -1), tf.float32)


def get_normalize_fn(quantile: List[float],
                     with_sqrt: bool,
                     with_z_score_normalization: bool,
                     eps: float = 1e-8,
                     preprocessing_fn: Optional[Callable[[types.Tensor],
                                                         types.Float]] = None):
  """Return a normalization function to normalize the input feature."""

  if not preprocessing_fn:
    # pylint: disable=unnecessary-lambda-assignment
    preprocessing_fn = lambda x: x
  processed_quantile = [preprocessing_fn(x) for x in quantile]
  mean = np.mean(processed_quantile)
  std = np.std(processed_quantile)

  def normalize(obs: types.Float):
    obs = tf.expand_dims(obs, -1)
    x = tf.cast(
        tf.raw_ops.Bucketize(input=obs, boundaries=quantile),
        tf.float32) / len(quantile)
    features = [x, x * x]
    if with_sqrt:
      features.append(tf.sqrt(x))
    if with_z_score_normalization:
      y = preprocessing_fn(tf.cast(obs, tf.float32))
      y = (y - mean) / (std + eps)
      features.append(y)
    return tf.concat(features, axis=-1)

  return normalize



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for blackbox_learner"""

import os
from absl.testing import absltest
import concurrent.futures
import gin
import tempfile
from typing import List
import numpy as np
import numpy.typing as npt
import tensorflow as tf
from tf_agents.networks import actor_distribution_network
from tf_agents.policies import actor_policy

from compiler_opt.distributed import worker
from compiler_opt.distributed.local import local_worker_manager
from compiler_opt.es import blackbox_learner, policy_utils
from compiler_opt.es import blackbox_optimizers
from compiler_opt.rl import corpus, inlining, policy_saver, registry
from compiler_opt.rl.inlining import config as inlining_config


@gin.configurable
class ESWorker(worker.Worker):
  """Temporary placeholder worker.
  Each time a worker is called, the function value
  it will return increases."""

  def __init__(self, arg, *, kwarg):
    self._arg = arg
    self._kwarg = kwarg
    self.function_value = 0.0

  def compile(self, policy: bytes, samples: List[corpus.ModuleSpec]) -> float:
    if policy and samples:
      self.function_value += 1.0
      return self.function_value
    else:
      return 0.0


class BlackboxLearnerTests(absltest.TestCase):
  """Tests for blackbox_learner"""

  def setUp(self):
    super().setUp()

    self._learner_config = blackbox_learner.BlackboxLearnerConfig(
        total_steps=1,
        blackbox_optimizer=blackbox_optimizers.Algorithm.MONTE_CARLO,
        est_type=blackbox_optimizers.EstimatorType.ANTITHETIC,
        fvalues_normalization=True,
        hyperparameters_update_method=blackbox_optimizers.UpdateMethod
        .NO_METHOD,
        num_top_directions=0,
        num_ir_repeats_within_worker=1,
        num_ir_repeats_across_worker=0,
        num_exact_evals=1,
        total_num_perturbations=3,
        precision_parameter=1,
        step_size=1.0)

    self._cps = corpus.create_corpus_for_testing(
        location=tempfile.gettempdir(),
        elements=[corpus.ModuleSpec(name='smth', size=1)],
        additional_flags=(),
        delete_flags=())

    output_dir = tempfile.gettempdir()
    policy_name = 'policy_name'

    # create a policy
    problem_config = registry.get_configuration(
        implementation=inlining.InliningConfig)
    time_step_spec, action_spec = problem_config.get_signature_spec()
    quantile_file_dir = os.path.join('compiler_opt', 'rl', 'inlining', 'vocab')
    creator = inlining_config.get_observation_processing_layer_creator(
        quantile_file_dir=quantile_file_dir,
        with_sqrt=False,
        with_z_score_normalization=False)
    layers = tf.nest.map_structure(creator, time_step_spec.observation)

    actor_network = actor_distribution_network.ActorDistributionNetwork(
        input_tensor_spec=time_step_spec.observation,
        output_tensor_spec=action_spec,
        preprocessing_layers=layers,
        preprocessing_combiner=tf.keras.layers.Concatenate(),
        fc_layer_params=(64, 64, 64, 64),
        dropout_layer_params=None,
        activation_fn=tf.keras.activations.relu)

    policy = actor_policy.ActorPolicy(
        time_step_spec=time_step_spec,
        action_spec=action_spec,
        actor_network=actor_network)

    # make the policy all zeros to be deterministic
    expected_policy_length = 17218
    policy_utils.set_vectorized_parameters_for_policy(policy, [0.0] *
                                                      expected_policy_length)
    init_params = policy_utils.get_vectorized_parameters_from_policy(policy)

    # save the policy
    saver = policy_saver.PolicySaver({policy_name: policy})
    policy_save_path = os.path.join(output_dir, 'temp_output', 'policy')
    saver.save(policy_save_path)

    def _policy_saver_fn(parameters: npt.NDArray[np.float32],
                         policy_name: str) -> None:
      if parameters is not None and policy_name:
        return None
      return None

    self._learner = blackbox_learner.BlackboxLearner(
        blackbox_opt=blackbox_optimizers.MonteCarloBlackboxOptimizer(
            precision_parameter=1.0,
            est_type=blackbox_optimizers.EstimatorType.ANTITHETIC,
            normalize_fvalues=True,
            hyperparameters_update_method=blackbox_optimizers.UpdateMethod
            .NO_METHOD,
            extra_params=None,
            step_size=1),
        sampler=self._cps,
        tf_policy_path=os.path.join(policy_save_path, policy_name),
        output_dir=output_dir,
        policy_saver_fn=_policy_saver_fn,
        model_weights=init_params,
        config=self._learner_config,
        seed=17)

  def test_get_perturbations(self):
    # test values generated with seed=17
    perturbations = self._learner._get_perturbations()  # pylint: disable=protected-access
    rng = np.random.default_rng(seed=17)
    for perturbation in perturbations:
      for value in perturbation:
        self.assertAlmostEqual(value, rng.normal())

  def test_get_results(self):
    with local_worker_manager.LocalWorkerPoolManager(
        ESWorker, count=3, arg='', kwarg='') as pool:
      self._samples = [[corpus.ModuleSpec(name='name1', size=1)],
                       [corpus.ModuleSpec(name='name2', size=1)],
                       [corpus.ModuleSpec(name='name3', size=1)]]
      perturbations = [b'00', b'01', b'10']
      # pylint: disable=protected-access
      results = self._learner._get_results(pool, perturbations)
      # pylint: enable=protected-access
      self.assertSequenceAlmostEqual([result.result() for result in results],
                                     [1.0, 1.0, 1.0])

  def test_get_rewards(self):
    f1 = concurrent.futures.Future()
    f1.set_exception(None)
    f2 = concurrent.futures.Future()
    f2.set_result(2)
    results = [f1, f2]
    rewards = self._learner._get_rewards(results)  # pylint: disable=protected-access
    self.assertEqual(rewards, [None, 2])

  def test_prune_skipped_perturbations(self):
    perturbations = [1, 2, 3, 4, 5]
    rewards = [1, None, 1, None, 1]
    blackbox_learner._prune_skipped_perturbations(perturbations, rewards)  # pylint: disable=protected-access
    self.assertListEqual(perturbations, [1, 3, 5])

  def test_run_step(self):
    with local_worker_manager.LocalWorkerPoolManager(
        ESWorker, count=3, arg='', kwarg='') as pool:
      self._learner.run_step(pool)  # pylint: disable=protected-access
      # expected length calculated from expected shapes of variables
      self.assertEqual(len(self._learner.get_model_weights()), 17218)
      # check that first 5 weights are not all zero
      # this will indicate general validity of all the values
      for value in self._learner.get_model_weights()[:5]:
        self.assertNotAlmostEqual(value, 0.0)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gymlike environment definition for MLGO."""

from __future__ import annotations
import dataclasses
from enum import Enum

import math
import subprocess
import abc
import contextlib
import io
import os
import tempfile
from typing import Callable, Generator, List, Optional, Tuple, Type

import numpy as np

from compiler_opt.rl import corpus
from compiler_opt.rl import log_reader


class StepType(Enum):
  FIRST = 1
  MID = 2
  LAST = 3


@dataclasses.dataclass
class TimeStep:
  obs: Optional[dict[str, np.NDArray]]
  reward: Optional[dict[str, float]]
  score_policy: Optional[dict[str, float]]
  score_default: Optional[dict[str, float]]
  context: Optional[str]
  module_name: str
  obs_id: Optional[int]
  step_type: StepType


_INTERACTIVE_PIPE_FILE_BASE = 'interactive-pipe-base'


class MLGOTask(metaclass=abc.ABCMeta):
  """Abstract base class for MLGO Tasks.

  A Task is an learning problem in LLVM, for example:
   - inlining-for-size
   - inlining-for-speed
   - register allocation (for speed)

  The Task type for a given problem defines how to build and score modules for
  the problem, both interactively and non-interactively.
  """

  @abc.abstractmethod
  def get_cmdline(self, clang_path: str, base_args: List[str],
                  interactive_base_path: Optional[str],
                  working_dir: str) -> List[str]:
    """Get the cmdline for building with this task.

    The resulting list[str] should be able to be passed to subprocess.run to
    execute clang.

    Args:
      clang_path: path to the clang executable.
      base_args: base arguments for building the module. Generally, these flags
        should not be modified and simply added to the result.
      interactive_base_path: the path to the interactive pipe base. if None,
        then don't run clang interactively.
      working_dir: directory where all artifacts from compilation should be
        written. This will be a temp directory whose lifetime is managed outside
        of the Task.

    Returns:
      The constructed command line.
    """
    pass

  @abc.abstractmethod
  def get_module_scores(self, working_dir: str) -> dict[str, float]:
    """Get the scores for each context in the module.

    This method should not be aware of whether the module was built with the
    default heuristic or a ML policy.

    Args:
      working_dir: Directory which was passed as working_dir to get_cmdline.
        Used to recover binaries/artifacts from the build

    Returns:
      A dictionary mapping [context name] -> [score].
    """
    pass


class ClangProcess:
  """Simple wrapper class around a clang process.

  This is used wrap both the clang process and the method to return the scores
  associated to the default-compiled binary.
  """

  def __init__(self, proc: subprocess.Popen,
               get_scores_fn: Callable[[], dict[str, float]], module_name):
    self._proc = proc
    self._get_scores_fn = get_scores_fn
    self._module_name = module_name

  def get_scores(self, timeout: Optional[int] = None):
    self._proc.wait(timeout=timeout)
    return self._get_scores_fn()


class InteractiveClang(ClangProcess):
  """Wrapper around clang's interactive mode."""

  def __init__(
      self,
      proc: subprocess.Popen,
      get_scores_fn: Callable[[], dict[str, float]],
      module_name: str,
      reader_pipe: io.BufferedReader,
      writer_pipe: io.BufferedWriter,
  ):
    super().__init__(proc, get_scores_fn, module_name)
    self._reader_pipe = reader_pipe
    self._writer_pipe = writer_pipe
    self._obs_gen = log_reader.read_log_from_file(self._reader_pipe)

    self._is_first_obs = True

    self._terminal_obs = TimeStep(
        obs=None,
        reward=None,
        score_policy=None,
        score_default=None,
        context=None,
        module_name=module_name,
        obs_id=None,
        step_type=StepType.LAST,
    )

  def _running(self) -> bool:
    return self._proc.poll() is None

  def get_observation(self) -> TimeStep:
    if not self._running():
      return self._terminal_obs

    def _get_step_type() -> StepType:
      step_type = StepType.FIRST if self._is_first_obs else StepType.MID
      self._is_first_obs = False
      return step_type

    try:
      obs: log_reader.ObservationRecord = next(self._obs_gen)

      tv_dict = {}
      for fv in obs.feature_values:
        array = fv.to_numpy()
        tv_dict[fv.spec.name] = np.reshape(array, newshape=fv.spec.shape)
      return TimeStep(
          obs=tv_dict,
          reward={obs.context: obs.score} if obs.score else None,
          score_policy=None,
          score_default=None,
          context=obs.context,
          module_name=self._module_name,
          obs_id=obs.observation_id,
          step_type=_get_step_type(),
      )
    except StopIteration:
      return self._terminal_obs

  def send_action(self, action: np.ndarray) -> None:
    assert self._running()
    data = action.tobytes()
    bytes_sent = self._writer_pipe.write(data)
    # Here we use the fact that for common types, the np.dtype and ctype should
    # behave the same
    assert bytes_sent == action.dtype.itemsize * math.prod(action.shape)
    try:
      self._writer_pipe.flush()
    except BrokenPipeError:
      # The pipe can break after we send the last action
      pass


_EPS = 1e-4


def compute_relative_rewards(score_a: dict[str, float],
                             score_b: dict[str, float]) -> dict[str, float]:

  def _reward_fn(a: float, b: float) -> float:
    return 1.0 - (a + _EPS) / (b + _EPS)

  assert score_a.keys() == score_b.keys()
  return {key: _reward_fn(score_a[key], score_b[key]) for key in score_a}


@contextlib.contextmanager
def clang_session(
    clang_path: str,
    module: corpus.LoadedModuleSpec,
    task_type: Type[MLGOTask],
    *,
    interactive: bool,
):
  """Context manager for clang session.

  We need to manage the context so resources like tempfiles and pipes have
  their lifetimes managed appropriately.

  Args:
    clang_path: The clang binary to use for the InteractiveClang session.
    module: The module to compile with clang.
    task_type: Type of the MLGOTask to use.
    interactive: Whether to use an interactive or default clang instance

  Yields:
    Either the constructed InteractiveClang or DefaultClang object.
  """
  with tempfile.TemporaryDirectory() as td:
    task_working_dir = os.path.join(td, '__task_working_dir__')
    os.mkdir(task_working_dir)
    task = task_type()

    base_args = list(module.build_command_line(td))
    interactive_base = os.path.join(
        td, _INTERACTIVE_PIPE_FILE_BASE) if interactive else None
    cmdline = task.get_cmdline(clang_path, base_args, interactive_base,
                               task_working_dir)

    def _get_scores() -> dict[str, float]:
      return task.get_module_scores(task_working_dir)

    writer_name = os.path.join(td, _INTERACTIVE_PIPE_FILE_BASE + '.in')
    reader_name = os.path.join(td, _INTERACTIVE_PIPE_FILE_BASE + '.out')
    if interactive:
      os.mkfifo(reader_name, 0o666)
      os.mkfifo(writer_name, 0o666)
    with subprocess.Popen(
        cmdline, stderr=subprocess.PIPE, stdout=subprocess.PIPE) as proc:
      try:
        if interactive:
          with io.BufferedWriter(io.FileIO(writer_name, 'wb')) as writer_pipe:
            with io.BufferedReader(io.FileIO(reader_name, 'rb')) as reader_pipe:
              yield InteractiveClang(
                  proc,
                  _get_scores,
                  module.name,
                  reader_pipe,
                  writer_pipe,
              )
        else:
          yield ClangProcess(
              proc,
              _get_scores,
              module.name,
          )

      finally:
        proc.kill()


def _get_clang_generator(
    clang_path: str,
    task_type: Type[MLGOTask],
) -> Generator[Optional[Tuple[ClangProcess, InteractiveClang]],
               Optional[corpus.LoadedModuleSpec], None]:
  """Returns a generator for creating InteractiveClang objects.

  TODO: fix this docstring

  Args:
    clang_path: Path to the clang binary to use within InteractiveClang.
    task_type: Type of the MLGO task to use.

  Returns:
    The generator for InteractiveClang objects.
  """
  while True:
    # The following line should be type-hinted as follows:
    #   module: corpus.LoadedModuleSpec = yield
    # However, this triggers a yapf crash. See:
    #   https://github.com/google/yapf/issues/1092
    module = yield
    with clang_session(
        clang_path, module, task_type, interactive=True) as iclang:
      with clang_session(
          clang_path, module, task_type, interactive=False) as clang:
        yield iclang, clang


class MLGOEnvironmentBase:
  """Base implementation for all MLGO environments.

  Depending on the RL framework, one may want different implementations of an
  enviroment (tf_agents: PyEnvironment, jax: dm-env, etc). This class
  implements the core methods that are needed to then implement any of these
  other environments as well.
  """

  def __init__(
      self,
      *,
      clang_path: str,
      task_type: Type[MLGOTask],
      obs_spec,
      action_spec,
  ):
    self._clang_generator = _get_clang_generator(clang_path, task_type)
    self._obs_spec = obs_spec
    self._action_spec = action_spec

    self._iclang: Optional[InteractiveClang] = None
    self._clang: Optional[ClangProcess] = None

  @property
  def obs_spec(self):
    return self._obs_spec

  @property
  def action_spec(self):
    return self._action_spec

  def observation(self):
    return self._last_obs

  def _get_observation(self) -> TimeStep:
    self._last_obs = self._iclang.get_observation()
    if self._last_obs.step_type == StepType.LAST:
      self._last_obs.score_policy = self._iclang.get_scores()
      self._last_obs.score_default = self._clang.get_scores()
      self._last_obs.reward = compute_relative_rewards(
          self._last_obs.score_policy, self._last_obs.score_default)
    return self._last_obs

  def reset(self, module: corpus.LoadedModuleSpec):
    # On the first call to reset(...), sending None starts the coroutine.
    # On subsequent calls, this resumes execution after
    # yielding the clang pair, which terminates the session pauses execution in
    # the coroutine where it awaits a module
    self._clang_generator.send(None)
    # pytype: disable=attribute-error
    self._iclang, self._clang = self._clang_generator.send(module)
    return self._get_observation()

  def step(self, action: np.ndarray):
    self._iclang.send_action(action)
    return self._get_observation()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for collecting data locally."""

import concurrent.futures
import itertools
import time
from typing import Callable, Dict, Iterator, List, Optional, Tuple

from absl import logging
from tf_agents.trajectories import trajectory

from compiler_opt.distributed import worker
from compiler_opt.distributed import buffered_scheduler
from compiler_opt.rl import best_trajectory
from compiler_opt.rl import compilation_runner
from compiler_opt.rl import corpus
from compiler_opt.rl import data_collector
from compiler_opt.rl import policy_saver


class LocalDataCollector(data_collector.DataCollector):
  """class for local data collection."""

  def __init__(
      self,
      cps: corpus.Corpus,
      num_modules: int,
      worker_pool: worker.WorkerPool,
      parser: Callable[[List[str]], Iterator[trajectory.Trajectory]],
      reward_stat_map: Dict[str, Optional[Dict[str,
                                               compilation_runner.RewardStat]]],
      best_trajectory_repo: Optional[best_trajectory.BestTrajectoryRepo],
      exit_checker_ctor=data_collector.EarlyExitChecker):
    # TODO(mtrofin): type exit_checker_ctor when we get typing.Protocol support
    super().__init__()

    self._corpus = cps
    self._num_modules = num_modules
    self._parser = parser
    self._worker_pool = worker_pool
    self._workers: List[
        compilation_runner
        .CompilationRunnerStub] = self._worker_pool.get_currently_active()
    self._reward_stat_map = reward_stat_map
    self._best_trajectory_repo = best_trajectory_repo
    self._exit_checker_ctor = exit_checker_ctor
    # _reset_workers is a future that resolves when post-data collection cleanup
    # work completes, i.e. cancelling all work and re-enabling the workers.
    # We remove this activity from the critical path by running it concurrently
    # with the training phase - i.e. whatever happens between successive data
    # collection calls. Subsequent runs will wait for these to finish.
    self._reset_workers: Optional[concurrent.futures.Future] = None
    self._current_futures: List[worker.WorkerFuture] = []
    self._pool = concurrent.futures.ThreadPoolExecutor()
    self._prefetch_pool = concurrent.futures.ThreadPoolExecutor()
    self._next_sample: List[
        concurrent.futures.Future] = self._prefetch_next_sample()

  def _prefetch_next_sample(self):
    t1 = time.time()
    sample = self._corpus.sample(k=self._num_modules, sort=False)
    ret = [
        self._prefetch_pool.submit(self._corpus.load_module_spec, element)
        for element in sample
    ]
    logging.info('prefetching took %d', time.time() - t1)
    return ret

  def close_pool(self):
    self._join_pending_jobs()
    # if the pool lost some workers, that's fine - we don't need to tell them
    # anything anymore. To the new ones, the call is redundant (fine).
    for p in self._workers:
      p.cancel_all_work()
    self._workers = None
    self._worker_pool = None

  def _join_pending_jobs(self):
    t1 = time.time()
    if self._reset_workers:
      concurrent.futures.wait([self._reset_workers])

    self._reset_workers = None
    # this should have taken negligible time, normally, since all the work
    # has been cancelled and the workers had time to process the cancellation
    # while training was unfolding.
    logging.info('Waiting for pending work from last iteration took %f',
                 time.time() - t1)

  def _schedule_jobs(self, policy: policy_saver.Policy, model_id: int,
                     sampled_modules: List[corpus.LoadedModuleSpec]) -> None:
    # by now, all the pending work, which was signaled to cancel, must've
    # finished
    self._join_pending_jobs()
    jobs = [
        dict(
            loaded_module_spec=loaded_module_spec,
            policy=policy,
            reward_stat=self._reward_stat_map[loaded_module_spec.name],
            model_id=model_id) for loaded_module_spec in sampled_modules
    ]

    (self._workers,
     self._current_futures) = buffered_scheduler.schedule_on_worker_pool(
         action=lambda w, kwargs: w.collect_data(**kwargs),
         jobs=jobs,
         worker_pool=self._worker_pool)

  def collect_data(
      self, policy: policy_saver.Policy, model_id: int
  ) -> Tuple[Iterator[trajectory.Trajectory], Dict[str, Dict[str, float]]]:
    """Collect data for a given policy.

    Args:
      policy: a policy_saver.Policy object to collect data with.

    Returns:
      An iterator of batched trajectory.Trajectory that are ready to be fed to
        training.
      A dict of extra monitoring information, e.g., how many modules succeeded.
      They will be reported using `tf.scalar.summary` by the trainer so these
      information is viewable in TensorBoard.
    """
    time1 = time.time()
    sampled_modules: List[corpus.LoadedModuleSpec] = [
        s.result() for s in self._next_sample
    ]
    logging.info('resolving prefetched sample took: %d seconds',
                 time.time() - time1)
    self._next_sample = self._prefetch_next_sample()
    self._schedule_jobs(policy, model_id, sampled_modules)

    def wait_for_termination():
      early_exit = self._exit_checker_ctor(num_modules=self._num_modules)

      def get_num_finished_work():
        finished_work = sum(res.done() for res in self._current_futures)
        return finished_work

      return early_exit.wait(get_num_finished_work)

    wait_seconds = wait_for_termination()
    current_work = list(zip(sampled_modules, self._current_futures))
    finished_work = [(spec, res) for spec, res in current_work if res.done()]
    successful_work = [(spec, res.result())
                       for spec, res in finished_work
                       if not worker.get_exception(res)]
    failures = len(finished_work) - len(successful_work)

    logging.info(('%d of %d modules finished in %d seconds (%d failures).'),
                 len(finished_work), self._num_modules, wait_seconds, failures)

    # signal whatever work is left to finish, and re-enable workers.
    def wrapup():
      cancel_futures = [wkr.cancel_all_work() for wkr in self._workers]
      worker.wait_for(cancel_futures)
      # now that the workers killed pending compilations, make sure the workers
      # drained their working queues first - they should all complete quickly
      # since the cancellation manager is killing immediately any process starts
      concurrent.futures.wait(self._current_futures)
      worker.wait_for([wkr.enable() for wkr in self._workers])

    self._reset_workers = self._pool.submit(wrapup)

    sequence_examples = list(
        itertools.chain.from_iterable(
            [res.serialized_sequence_examples for (_, res) in successful_work]))
    total_trajectory_length = sum(res.length for (_, res) in successful_work)
    self._reward_stat_map.update(
        {spec.name: res.reward_stats for (spec, res) in successful_work})

    if self._best_trajectory_repo is not None:
      for spec, res in successful_work:
        module_name = spec.name
        for (identifier, reward,
             sequence_example) in zip(res.keys, res.policy_rewards,
                                      res.serialized_sequence_examples):
          self._best_trajectory_repo.update_if_better_trajectory(
              module_name, identifier, reward, sequence_example)

    monitor_dict = {}
    monitor_dict['default'] = {
        'success_modules': len(successful_work),
        'total_trajectory_length': total_trajectory_length,
    }
    rewards = list(
        itertools.chain.from_iterable(
            [res.rewards for (_, res) in successful_work]))
    monitor_dict[
        'reward_distribution'] = data_collector.build_distribution_monitor(
            rewards)

    parsed = self._parser(sequence_examples)

    return parsed, monitor_dict

  def on_dataset_consumed(self,
                          dataset_iterator: Iterator[trajectory.Trajectory]):
    pass



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""util function to create a tf_agent."""

from typing import Any, Callable, Dict

import abc
import gin
import tensorflow as tf

from tf_agents.agents import tf_agent
from tf_agents.agents.behavioral_cloning import behavioral_cloning_agent
from tf_agents.agents.dqn import dqn_agent
from tf_agents.agents.ppo import ppo_agent
from tf_agents.specs import tensor_spec
from tf_agents.typing import types

from compiler_opt.rl import constant_value_network
from compiler_opt.rl.distributed import agent as distributed_ppo_agent


class AgentConfig(metaclass=abc.ABCMeta):
  """Agent creation and data processing hook-ups."""

  def __init__(self, *, time_step_spec: types.NestedTensorSpec,
               action_spec: types.NestedTensorSpec):
    self._time_step_spec = time_step_spec
    self._action_spec = action_spec

  @property
  def time_step_spec(self):
    return self._time_step_spec

  @property
  def action_spec(self):
    return self._action_spec

  @abc.abstractmethod
  def create_agent(self, preprocessing_layers: tf.keras.layers.Layer,
                   policy_network: types.Network) -> tf_agent.TFAgent:
    """Specific agent configs must implement this."""
    raise NotImplementedError()

  def get_policy_info_parsing_dict(
      self) -> Dict[str, tf.io.FixedLenSequenceFeature]:
    """Return the parsing dict for the policy info."""
    return {}

  # pylint: disable=unused-argument
  def process_parsed_sequence_and_get_policy_info(
      self, parsed_sequence: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
    """Function to process parsed_sequence and to return policy_info.

    Args:
      parsed_sequence: A dict from feature_name to feature_value parsed from TF
        SequenceExample.

    Returns:
      A nested policy_info for given agent.
    """
    return {}


@gin.configurable
def create_agent(agent_config: AgentConfig,
                 preprocessing_layer_creator: Callable[[types.TensorSpec],
                                                       tf.keras.layers.Layer],
                 policy_network: types.Network):
  """Gin configurable wrapper of AgentConfig.create_agent.
  Works around the fact that class members aren't gin-configurable."""
  preprocessing_layers = tf.nest.map_structure(
      preprocessing_layer_creator, agent_config.time_step_spec.observation)
  return agent_config.create_agent(preprocessing_layers, policy_network)


@gin.configurable(module='agents')
class BCAgentConfig(AgentConfig):
  """Behavioral Cloning agent configuration."""

  def create_agent(self, preprocessing_layers: tf.keras.layers.Layer,
                   policy_network: types.Network) -> tf_agent.TFAgent:
    """Creates a behavioral_cloning_agent."""

    network = policy_network(
        self.time_step_spec.observation,
        self.action_spec,
        preprocessing_layers=preprocessing_layers,
        name='QNetwork')

    return behavioral_cloning_agent.BehavioralCloningAgent(
        self.time_step_spec,
        self.action_spec,
        cloning_network=network,
        num_outer_dims=2)


@gin.configurable(module='agents')
class DQNAgentConfig(AgentConfig):
  """DQN agent configuration."""

  def create_agent(self, preprocessing_layers: tf.keras.layers.Layer,
                   policy_network: types.Network) -> tf_agent.TFAgent:
    """Creates a dqn_agent."""
    network = policy_network(
        self.time_step_spec.observation,
        self.action_spec,
        preprocessing_layers=preprocessing_layers,
        name='QNetwork')

    return dqn_agent.DqnAgent(
        self.time_step_spec, self.action_spec, q_network=network)


@gin.configurable(module='agents')
class PPOAgentConfig(AgentConfig):
  """PPO/Reinforce agent configuration."""

  def create_agent(self, preprocessing_layers: tf.keras.layers.Layer,
                   policy_network: types.Network) -> tf_agent.TFAgent:
    """Creates a ppo_agent."""

    actor_network = policy_network(
        self.time_step_spec.observation,
        self.action_spec,
        preprocessing_layers=preprocessing_layers,
        name='ActorDistributionNetwork')

    critic_network = constant_value_network.ConstantValueNetwork(
        self.time_step_spec.observation, name='ConstantValueNetwork')

    return ppo_agent.PPOAgent(
        self.time_step_spec,
        self.action_spec,
        actor_net=actor_network,
        value_net=critic_network)

  def get_policy_info_parsing_dict(
      self) -> Dict[str, tf.io.FixedLenSequenceFeature]:
    if tensor_spec.is_discrete(self._action_spec):
      return {
          'CategoricalProjectionNetwork_logits':
              tf.io.FixedLenSequenceFeature(
                  shape=(self._action_spec.maximum - self._action_spec.minimum +
                         1),
                  dtype=tf.float32)
      }
    else:
      return {
          'NormalProjectionNetwork_scale':
              tf.io.FixedLenSequenceFeature(shape=(), dtype=tf.float32),
          'NormalProjectionNetwork_loc':
              tf.io.FixedLenSequenceFeature(shape=(), dtype=tf.float32)
      }

  def process_parsed_sequence_and_get_policy_info(
      self, parsed_sequence: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
    if tensor_spec.is_discrete(self._action_spec):
      policy_info = {
          'dist_params': {
              'logits': parsed_sequence['CategoricalProjectionNetwork_logits']
          }
      }
      del parsed_sequence['CategoricalProjectionNetwork_logits']
    else:
      policy_info = {
          'dist_params': {
              'scale': parsed_sequence['NormalProjectionNetwork_scale'],
              'loc': parsed_sequence['NormalProjectionNetwork_loc']
          }
      }
      del parsed_sequence['NormalProjectionNetwork_scale']
      del parsed_sequence['NormalProjectionNetwork_loc']
    return policy_info


@gin.configurable(module='agents')
class DistributedPPOAgentConfig(PPOAgentConfig):
  """Distributed PPO/Reinforce agent configuration."""

  def _create_agent_implt(self, preprocessing_layers: tf.keras.layers.Layer,
                          policy_network: types.Network) -> tf_agent.TFAgent:
    """Creates a ppo_distributed agent."""
    actor_network = policy_network(
        self.time_step_spec.observation,
        self.action_spec,
        preprocessing_layers=preprocessing_layers,
        preprocessing_combiner=tf.keras.layers.Concatenate(),
        name='ActorDistributionNetwork')

    critic_network = constant_value_network.ConstantValueNetwork(
        self.time_step_spec.observation, name='ConstantValueNetwork')

    return distributed_ppo_agent.MLGOPPOAgent(
        self.time_step_spec,
        self.action_spec,
        optimizer=tf.keras.optimizers.Adam(learning_rate=4e-4, epsilon=1e-5),
        actor_net=actor_network,
        value_net=critic_network,
        value_pred_loss_coef=0.0,
        entropy_regularization=0.01,
        importance_ratio_clipping=0.2,
        discount_factor=1.0,
        gradient_clipping=1.0,
        debug_summaries=False,
        value_clipping=None,
        aggregate_losses_across_replicas=True,
        loss_scaling_factor=1.0)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compilation problem component model registry.

This allows tools just get a ProblemConfiguration object encapsulating all
necessary elements - compilation runner implementation & ML configuration:

`config = registry.get_configuration()`

See also problem_configuration.py
"""

from typing import Type

import gin
import tf_agents as tfa

from compiler_opt.rl import problem_configuration

# Register implementations. They appear unused, but they need to be imported
# to trigger gin registration.
import compiler_opt.rl.inlining  # pylint: disable=unused-import
import compiler_opt.rl.regalloc  # pylint: disable=unused-import

types = tfa.typing.types


@gin.configurable(module='config_registry')
def get_configuration(
    implementation: Type[problem_configuration.ProblemConfiguration]
) -> problem_configuration.ProblemConfiguration:
  return implementation()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Reader for the simple log format (non-protobuf).

The "simple" log format consists of a sequence of one-line json strings and
raw dump of tensor data - with the assumption that the endianness of the
reader is the same as the endianness of the producer - as follows:

header: this is a json object describing the feature tensors and the outcome
tensor. The feature tensors are in an array - the order matters (as will become
apparent).
Example: {"features": [{tensor spec}, {tensor spec}], "score": {tensor spec}}

The tensor spec is a json object:
{"name":.., "port":... "shape":[..], "type":".."}

context: this is a json object indicating the context the observations that
follow refer to. Example: for inlining, the context is "default" (the module).
For regalloc, the context is a function name.
Example: {"context": "_ZNfoobar"}

observation: this is a json object indicating that an observation - i.e. feature
data - is following. It also contains the observation count (0, 1...)
Example: {"observation": 0}

A buffer containing the dump of tensor data, in the order given in the header,
follows here. A new line terminates it - just so that the next json string
appears at the beginning of the line, in case the log is opened with an editor
(so just for debugging). The reader should use the header data to know how much
data to read, and to which tensors it corresponds. The reader may not rely on
the terminating \n as indicator, and should just assume it there and consume it
upon finishing reading the tensor data.

outcome: this is a json object indicating the score/reward data follows. It also
has an id which should match that of the observation before (for debugging)
Example: {"outcome": 0}

A buffer containing the outcome tensor follows - same idea as for features.

The above repeats - either a new observation follows, or a new context.

Refer to (in llvm repo) to llvm/lib/Analysis/models/log_reader.py
which is used there for testing.
"""

import collections
import ctypes
import dataclasses
import json
import math

from compiler_opt import type_map
from typing import Any, BinaryIO, Dict, Generator, List, Optional
import numpy as np
import tensorflow as tf

_element_type_name_to_dtype = {
    name: dtype for name, _, dtype in type_map.TYPE_ASSOCIATIONS
}

_dtype_to_ctype = {
    dtype: ctype for _, ctype, dtype in type_map.TYPE_ASSOCIATIONS
}


def create_tensorspec(d: Dict[str, Any]) -> tf.TensorSpec:
  name: str = d['name']
  shape: List[int] = [int(e) for e in d['shape']]
  element_type_str: str = d['type']
  if element_type_str not in _element_type_name_to_dtype:
    raise ValueError(f'uknown type: {element_type_str}')
  return tf.TensorSpec(
      name=name,
      shape=tf.TensorShape(shape),
      dtype=_element_type_name_to_dtype[element_type_str])


class LogReaderTensorValue:
  """The value of a tensor of a given spec.

  We root the bytes buffer which provide the underlying data, and index in
  the value based on the type of the tensor, thus the TensorValue can be used
  as a list-like object containing the scalar values, in row-major order, of
  the tensor.

  Endianness is assumed to be the same as the log producer's.
  """
  __slots__ = ('_buffer', '_spec', '_len', '_view')

  def __init__(self, spec: tf.TensorSpec, buffer: bytes):
    self._buffer = buffer
    self._spec = spec
    self._len = math.prod(spec.shape)
    self._set_view()

  @property
  def spec(self):
    return self._spec

  def to_numpy(self) -> np.ndarray:
    return np.frombuffer(
        self._buffer, dtype=_dtype_to_ctype[self._spec.dtype], count=self._len)

  def _set_view(self):
    # c_char_p is a nul-terminated string, so the more appropriate cast here
    # would be POINTER(c_char), but unfortunately, c_char_p is the only
    # type that can be constructed from a `bytes`. To capture our intent,
    # we cast the c_char_p to
    buffer_as_nul_ending_ptr = ctypes.c_char_p(self._buffer)
    buffer_as_naked_ptr = ctypes.cast(buffer_as_nul_ending_ptr,
                                      ctypes.POINTER(ctypes.c_char))
    self._view = ctypes.cast(buffer_as_naked_ptr,
                             ctypes.POINTER(_dtype_to_ctype[self.spec.dtype]))

  def __len__(self) -> int:
    return self._len

  def __getitem__(self, index: int):
    if index < 0 or index >= self._len:
      raise IndexError(f'Index {index} out of range [0..{self._len})')
    # pytype believes `index` is an object, despite all the annotations to the
    # contrary.
    return self._view[index]  # pytype:disable=wrong-arg-types


@dataclasses.dataclass(frozen=True)
class _Header:
  features: List[tf.TensorSpec]
  score: Optional[tf.TensorSpec]


def _read_tensor(fs: BinaryIO, ts: tf.TensorSpec) -> LogReaderTensorValue:
  size = math.prod(ts.shape) * ctypes.sizeof(_dtype_to_ctype[ts.dtype])
  data = fs.read(size)
  return LogReaderTensorValue(ts, data)


def _read_header(f: BinaryIO) -> Optional[_Header]:
  header_raw = f.readline()
  if not header_raw:
    # This is the path taken by empty files
    return None
  header = json.loads(header_raw)
  tensor_specs = [create_tensorspec(ts) for ts in header['features']]
  score_spec = create_tensorspec(header['score']) if 'score' in header else None
  return _Header(features=tensor_specs, score=score_spec)


@dataclasses.dataclass(frozen=True)
class ObservationRecord:
  context: str
  observation_id: int
  feature_values: List[LogReaderTensorValue]
  score: Optional[LogReaderTensorValue]


def _enumerate_log_from_stream(
    f: BinaryIO, header: _Header) -> Generator[ObservationRecord, None, None]:
  """A generator that returns Record objects from a log file.

  It is assumed the log file's header was read separately.
  """
  tensor_specs = header.features
  score_spec = header.score
  context = None
  while event_str := f.readline():
    event = json.loads(event_str)
    if 'context' in event:
      context = event['context']
      continue
    observation_id = int(event['observation'])
    features = []
    for ts in tensor_specs:
      features.append(_read_tensor(f, ts))
    f.readline()
    score = None
    if score_spec is not None:
      score_header = json.loads(f.readline())
      assert int(score_header['outcome']) == observation_id
      score = _read_tensor(f, score_spec)
      f.readline()
    yield ObservationRecord(
        context=context,
        observation_id=observation_id,
        feature_values=features,
        score=score)


def read_log_from_file(f) -> Generator[ObservationRecord, None, None]:
  header = _read_header(f)
  if header:
    yield from _enumerate_log_from_stream(f, header)


def read_log(fname: str) -> Generator[ObservationRecord, None, None]:
  with open(fname, 'rb') as f:
    yield from read_log_from_file(f)


def _add_feature(se: tf.train.SequenceExample, spec: tf.TensorSpec,
                 value: LogReaderTensorValue):
  f = se.feature_lists.feature_list[spec.name].feature.add()
  # This should never happen: _add_feature is an implementation detail of
  # read_log_as_sequence_examples, and the only dtypes we should see here are
  # those in _element_type_name_map, or an exception would have been thrown
  # already.
  if spec.dtype not in _dtype_to_ctype:
    raise ValueError('Unsupported dtype: f{spec.dtype}')
  if spec.dtype in [tf.float32, tf.float64]:
    lst = f.float_list.value
  else:
    lst = f.int64_list.value
  lst.extend(value)


def read_log_as_sequence_examples(
    fname: str) -> Dict[str, tf.train.SequenceExample]:
  ret: Dict[str, tf.train.SequenceExample] = collections.defaultdict(
      tf.train.SequenceExample)
  # a record is an observation: the features and score for one step.
  # the records are in time order
  # the `context` is, for example, the function name for passes like regalloc.
  # we produce a dictionary keyed in contexts with SequenceExample values.
  for record in read_log(fname):
    se = ret[record.context]
    if record.score:
      _add_feature(se, record.score.spec, record.score)
    for t in record.feature_values:
      _add_feature(se, t.spec, t)
  return ret



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A script for filtering gtests based on whether or not they pass/fail

Within Chromium specifically, there are some test executables that have a lot
of tests that are excellent for benchmarking, but running these test suites
in their entirety can sometimes be problematic as some subsets of the tests
available in the executable might require certain hardware configurations
such as an X configuration with working graphics drivers, and we would prefer
to avoid those tests. This exectuable goes through an entire test suite
description and returns another test suite description containing only tests
that pass.

Usage:
PYTHONPATH=$PYTHONPATH:. python3 \
  ./compiler_opt/benchmark/filter_tests.py \
  --input_tests=./compiler_opt/benchmark/chromium_test_descriptions \
    /browser_tests.json \
  --output_tests=./browser_tests_filtered.json \
  --num_threads=32 \
  --executable_path=/chromium/src/out/Release/browser_tests
"""

import json
import os

from absl import flags
from absl import app
from absl import logging

from compiler_opt.benchmark import gtest_executable_utils

FLAGS = flags.FLAGS

flags.DEFINE_string('input_tests', '',
                    'The path to the test description JSON to filter.')
flags.DEFINE_string(
    'output_tests', '',
    'The path to the JSON file to place the output test suite '
    'description.')
flags.DEFINE_integer(
    'num_threads', 1, 'The number of threads to use for running tests in '
    'parallel.')
flags.DEFINE_string(
    'executable_path', '',
    'The path to the Chromium build directory where all the '
    'test executables are stored')


def main(_):
  if not os.path.exists(FLAGS.executable_path):
    logging.fatal('Executable path does not exist.')
  with open(FLAGS.input_tests, encoding='UTF-8') as test_description_file:
    test_suite_description = json.load(test_description_file)
    test_outputs = gtest_executable_utils.run_test_suite(
        test_suite_description, FLAGS.executable_path, [], FLAGS.num_threads)
    test_list = []
    for test_output in test_outputs:
      test_list.append(test_output['name'])
    # copy the old test suite and just replace the tests array
    new_test_suite_description = test_suite_description
    new_test_suite_description['tests'] = test_list
    with open(FLAGS.output_tests, 'w', encoding='UTF-8') as tests_output_file:
      json.dump(new_test_suite_description, tests_output_file)


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compiler_opt.rl.best_trajectory."""

from absl.testing import absltest
from absl.testing import parameterized

import tensorflow as tf

from compiler_opt.rl import best_trajectory

_ACTION_NAME = 'mock'


def _get_test_repo_1():
  repo = best_trajectory.BestTrajectoryRepo(action_name=_ACTION_NAME)
  # pylint: disable=protected-access
  repo._best_trajectories['module_1'] = {
      'function_1':
          best_trajectory.BestTrajectory(reward=3.4, action_list=[1, 3, 5]),
      'function_2':
          best_trajectory.BestTrajectory(reward=1.2, action_list=[9, 7, 5])
  }
  # pylint: enable=protected-access
  return repo


def _get_test_repo_2():
  repo = best_trajectory.BestTrajectoryRepo(action_name=_ACTION_NAME)
  # pylint: disable=protected-access
  repo._best_trajectories['module_1'] = {
      'function_1':
          best_trajectory.BestTrajectory(reward=2.3, action_list=[1, 3]),
      'function_2':
          best_trajectory.BestTrajectory(reward=3.4, action_list=[9, 7])
  }
  repo._best_trajectories['module_2'] = {
      'function_1':
          best_trajectory.BestTrajectory(reward=7.8, action_list=[2, 4, 6]),
  }
  # pylint: enable=protected-access
  return repo


def _get_combined_repo():
  repo = best_trajectory.BestTrajectoryRepo(action_name=_ACTION_NAME)
  # pylint: disable=protected-access
  repo._best_trajectories['module_1'] = {
      'function_1':
          best_trajectory.BestTrajectory(reward=2.3, action_list=[1, 3]),
      'function_2':
          best_trajectory.BestTrajectory(reward=1.2, action_list=[9, 7, 5])
  }
  repo._best_trajectories['module_2'] = {
      'function_1':
          best_trajectory.BestTrajectory(reward=7.8, action_list=[2, 4, 6]),
  }
  # pylint: enable=protected-access
  return repo


def _create_sequence_example(action_list):
  example = tf.train.SequenceExample()
  for action in action_list:
    example.feature_lists.feature_list[_ACTION_NAME].feature.add(
    ).int64_list.value.append(action)
  return example.SerializeToString()


class BestTrajectoryTest(parameterized.TestCase):

  @parameterized.named_parameters(('repo_1', _get_test_repo_1()),
                                  ('repo_2', _get_test_repo_2()))
  def test_sink_load_json_file(self, repo):
    path = self.create_tempfile().full_path
    repo.sink_to_json_file(path)
    loaded_repo = best_trajectory.BestTrajectoryRepo(action_name=_ACTION_NAME)
    loaded_repo.load_from_json_file(path)
    self.assertDictEqual(repo.best_trajectories, loaded_repo.best_trajectories)

  def test_sink_to_csv_file(self):
    path = self.create_tempfile().full_path
    repo = _get_test_repo_1()
    repo.sink_to_csv_file(path)
    with open(path, 'r', encoding='utf-8') as f:
      text = f.read()

    self.assertEqual(text,
                     'module_1,function_1,1,3,5\nmodule_1,function_2,9,7,5\n')

  @parameterized.named_parameters(
      {
          'testcase_name': 'repo_1_combine_2',
          'base_repo': _get_test_repo_1(),
          'second_repo': _get_test_repo_2()
      }, {
          'testcase_name': 'repo_2_combine_1',
          'base_repo': _get_test_repo_2(),
          'second_repo': _get_test_repo_1()
      })
  def test_combine_with_other_repo(self, base_repo, second_repo):
    base_repo.combine_with_other_repo(second_repo)
    self.assertDictEqual(base_repo.best_trajectories,
                         _get_combined_repo().best_trajectories)

  def test_update_if_better_trajectory(self):
    repo = _get_test_repo_1()
    repo.update_if_better_trajectory(
        'module_1', 'function_1', 2.3,
        _create_sequence_example(action_list=[1, 3]))
    repo.update_if_better_trajectory(
        'module_1', 'function_2', 3.4,
        _create_sequence_example(action_list=[9, 7]))
    repo.update_if_better_trajectory(
        'module_2', 'function_1', 7.8,
        _create_sequence_example(action_list=[2, 4, 6]))
    self.assertDictEqual(repo.best_trajectories,
                         _get_combined_repo().best_trajectories)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Register allocation training config."""

import gin
import tensorflow as tf
from tf_agents.specs import tensor_spec
from tf_agents.trajectories import time_step
from compiler_opt.rl import feature_ops


def get_num_registers():
  return 33


# pylint: disable=g-complex-comprehension
@gin.configurable()
def get_regalloc_signature_spec():
  """Returns (time_step_spec, action_spec) for LLVM register allocation."""
  num_registers = get_num_registers()

  observation_spec = dict(
      (key, tf.TensorSpec(dtype=tf.int64, shape=(num_registers), name=key))
      for key in ('mask', 'is_hint', 'is_local', 'is_free'))
  observation_spec.update(
      dict((key,
            tensor_spec.BoundedTensorSpec(
                dtype=tf.int64,
                shape=(num_registers),
                name=key,
                minimum=0,
                maximum=6)) for key in ('max_stage', 'min_stage')))
  observation_spec.update(
      dict((key,
            tf.TensorSpec(dtype=tf.float32, shape=(num_registers), name=key))
           for key in ('weighed_reads_by_max', 'weighed_writes_by_max',
                       'weighed_read_writes_by_max', 'weighed_indvars_by_max',
                       'hint_weights_by_max', 'start_bb_freq_by_max',
                       'end_bb_freq_by_max', 'hottest_bb_freq_by_max',
                       'liverange_size', 'use_def_density', 'nr_defs_and_uses',
                       'nr_broken_hints', 'nr_urgent', 'nr_rematerializable')))
  observation_spec['progress'] = tensor_spec.BoundedTensorSpec(
      dtype=tf.float32, shape=(), name='progress', minimum=0, maximum=1)

  reward_spec = tf.TensorSpec(dtype=tf.float32, shape=(), name='reward')
  time_step_spec = time_step.time_step_spec(observation_spec, reward_spec)

  action_spec = tensor_spec.BoundedTensorSpec(
      dtype=tf.int64,
      shape=(),
      name='index_to_evict',
      minimum=0,
      maximum=num_registers - 1)

  return time_step_spec, action_spec


@gin.configurable
def get_observation_processing_layer_creator(quantile_file_dir=None,
                                             with_sqrt=True,
                                             with_z_score_normalization=True,
                                             eps=1e-8):
  """Wrapper for observation_processing_layer."""
  quantile_map = feature_ops.build_quantile_map(quantile_file_dir)

  def observation_processing_layer(obs_spec):
    """Creates the layer to process observation given obs_spec."""
    if obs_spec.name in ('mask', 'nr_urgent'):
      return tf.keras.layers.Lambda(feature_ops.discard_fn)

    if obs_spec.name in ('is_hint', 'is_local', 'is_free'):
      return tf.keras.layers.Lambda(feature_ops.identity_fn)

    if obs_spec.name in ('max_stage', 'min_stage'):
      return tf.keras.layers.Embedding(7, 4)

    normalize_fn = log_normalize_fn = None
    if obs_spec.name not in get_nonnormalized_features():
      quantile = quantile_map[obs_spec.name]

      first_non_zero = 0
      for x in quantile:
        if x > 0:
          first_non_zero = x
          break

      normalize_fn = feature_ops.get_normalize_fn(quantile, with_sqrt,
                                                  with_z_score_normalization,
                                                  eps)
      log_normalize_fn = feature_ops.get_normalize_fn(
          quantile,
          with_sqrt,
          with_z_score_normalization,
          eps,
          preprocessing_fn=lambda x: tf.math.log(x + first_non_zero))

    if obs_spec.name in ['nr_rematerializable', 'nr_broken_hints']:
      return tf.keras.layers.Lambda(normalize_fn)

    if obs_spec.name in ('liverange_size', 'nr_defs_and_uses'
                        ) or obs_spec.name.endswith('by_max'):
      return tf.keras.layers.Lambda(log_normalize_fn)

    if obs_spec.name == 'use_def_density':

      def use_def_density_processing_fn(obs):
        features = [tf.where(tf.math.is_inf(tf.expand_dims(obs, -1)), 1.0, 0.0)]
        obs = tf.where(tf.math.is_inf(obs), 1.0, obs)
        features.append(log_normalize_fn(obs))
        # pylint: disable=unexpected-keyword-arg
        return tf.concat(features, axis=-1)

      return tf.keras.layers.Lambda(use_def_density_processing_fn)

    if obs_spec.name == 'progress':

      def progress_processing_fn(obs):
        obs = tf.expand_dims(obs, -1)
        obs = tf.tile(obs, [1, get_num_registers()])
        obs = normalize_fn(obs)
        return obs

      return tf.keras.layers.Lambda(progress_processing_fn)

    # Make sure all features have a preprocessing function.
    raise KeyError('Missing preprocessing function for some feature.')

  return observation_processing_layer


def get_nonnormalized_features():
  return [
      'mask', 'nr_urgent', 'is_hint', 'is_local', 'is_free', 'max_stage',
      'min_stage', 'reward'
  ]



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library to train a policy with PPO"""

import time
import os

from absl import logging

import gin
import reverb
import tensorflow.compat.v2 as tf

from tf_agents.experimental.distributed import reverb_variable_container
from tf_agents.replay_buffers import reverb_replay_buffer
from tf_agents.train import learner as actor_learner
from tf_agents.train import triggers
from tf_agents.utils import common

from compiler_opt.rl import gin_external_configurables  # pylint: disable=unused-import
from compiler_opt.rl import agent_config
from compiler_opt.rl import registry
from compiler_opt.rl.distributed import learner as learner_lib

_SHUFFLE_BUFFER_EPISODE_LEN = 3


@gin.configurable
def train(
    root_dir: str,
    strategy: tf.distribute.Strategy,
    replay_buffer_server_address: str,
    variable_container_server_address: str,
    # Training params
    per_replica_batch_size,
    num_epochs: int,
    num_iterations: int,
    num_episodes_per_iteration: int,
    sequence_length: int,
    summary_interval: int):
  """Trains a PPO agent."""
  # Get the specs from the environment.
  problem_config = registry.get_configuration()
  time_step_spec, action_spec = problem_config.get_signature_spec()

  # Create the agent.
  with strategy.scope():
    train_step = tf.compat.v1.train.get_or_create_global_step()
    agent = agent_config.create_agent(
        agent_config.DistributedPPOAgentConfig(
            time_step_spec=time_step_spec, action_spec=action_spec),
        preprocessing_layer_creator=problem_config
        .get_preprocessing_layer_creator())
    model_id = common.create_variable('model_id')
    # The model_id should equal to the iteration number.
    model_id.assign(0)
    agent.initialize()

  # Create the policy saver which saves the initial model now, then it
  # periodically checkpoints the policy weigths.
  saved_model_dir = os.path.join(root_dir, actor_learner.POLICY_SAVED_MODEL_DIR)
  save_model_trigger = triggers.PolicySavedModelTrigger(
      saved_model_dir, agent, train_step, interval=1000)

  # Create the variable container.
  variables = {
      reverb_variable_container.POLICY_KEY: agent.collect_policy.variables(),
      reverb_variable_container.TRAIN_STEP_KEY: train_step,
      'model_id': model_id,
  }
  variable_container = reverb_variable_container.ReverbVariableContainer(
      variable_container_server_address,
      table_names=[reverb_variable_container.DEFAULT_TABLE])
  variable_container.push(variables)

  # Create the replay buffer.
  reverb_replay_train = reverb_replay_buffer.ReverbReplayBuffer(
      agent.collect_data_spec,
      sequence_length=sequence_length,
      table_name='training_table',
      server_address=replay_buffer_server_address)

  # Initialize the dataset.
  def experience_dataset_fn():
    get_dtype = lambda x: x.dtype  # pylint: disable=unnecessary-lambda-assignment
    get_shape = lambda x: (None,) + x.shape  # pylint: disable=unnecessary-lambda-assignment

    shapes = tf.nest.map_structure(get_shape, agent.collect_data_spec)
    dtypes = tf.nest.map_structure(get_dtype, agent.collect_data_spec)

    dataset = reverb.TrajectoryDataset(
        server_address=replay_buffer_server_address,
        table='training_table',
        dtypes=dtypes,
        shapes=shapes,
        # Menger uses learner_iterations_per_call (256). Using 8 here instead
        # because we do not need that much data in the buffer (they have to be
        # filtered out for the next iteration anyways). The rule of thumb is
        # 2-3x batch_size.
        max_in_flight_samples_per_worker=8,
        num_workers_per_iterator=-1,
        max_samples_per_stream=-1,
        rate_limiter_timeout_ms=-1,
    )

    def broadcast_info(info_traj):
      # Assumes that the first element of traj is shaped
      # (sequence_length, ...); and we extract this length.
      info, traj = info_traj
      first_elem = tf.nest.flatten(traj)[0]
      length = first_elem.shape[0] or tf.shape(first_elem)[0]
      info = tf.nest.map_structure(lambda t: tf.repeat(t, [length]), info)
      return reverb.ReplaySample(info, traj)

    dataset = dataset.map(broadcast_info)
    return dataset

  # Create the learner.
  learning_triggers = [
      save_model_trigger,
      triggers.StepPerSecondLogTrigger(train_step, interval=100)
  ]

  def per_sequence_fn(sample):
    # At this point, each sample data contains a sequence of trajectories.
    data, info = sample.data, sample.info
    data = agent.preprocess_sequence(data)
    return data, info

  learner = learner_lib.MLGOPPOLearner(
      root_dir,
      train_step,
      model_id,
      agent,
      experience_dataset_fn,
      sequence_length,
      num_episodes_per_iteration=num_episodes_per_iteration,
      minibatch_size=per_replica_batch_size,
      shuffle_buffer_size=(_SHUFFLE_BUFFER_EPISODE_LEN * sequence_length),
      triggers=learning_triggers,
      summary_interval=summary_interval,
      strategy=strategy,
      num_epochs=num_epochs,
      per_sequence_fn=per_sequence_fn,
      allow_variable_length_episodes=False)

  # Run the training loop.
  for i in range(num_iterations):
    step_val = train_step.numpy()
    logging.info('Training. Iteration: %d', i)
    start_time = time.time()
    learner.wait_for_data()
    data_wait_time = time.time() - start_time
    logging.info('Data wait time sec: %s', data_wait_time)
    loss_info = learner.run()
    logging.info('Loss info: %s', str(loss_info))
    num_steps = train_step.numpy() - step_val
    run_time = time.time() - start_time
    logging.info('Steps per sec: %s', num_steps / run_time)
    logging.info('Pushing variables at model_id: %d', model_id.numpy())
    variable_container.push(variables)
    logging.info('clearing replay buffer')
    reverb_replay_train.clear()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Convert benchmark results.json to csv.

To run:
python3 \
compiler_opt/benchmark/benchmark_report_converter.py \
  --base=/tmp/base_report.json \
  --exp=/tmp/exp_report.json \
  --counters=INSTRUCTIONS \
  --counters=CYCLES \
  --output=/tmp/summary.csv

optionally, add --suite_name=<name of benchmark>, if batch-processing multiple
benchmarks' reports.

Assuming /tmp/{base|exp}_report.json were produced from benchmark runs, which
were asked to collect the counters named INSTRUCTIONS and CYCLES.
"""

import csv
import json

from typing import Sequence

from absl import app
from absl import flags

import tensorflow.compat.v2 as tf

from compiler_opt.benchmark import benchmark_report

flags.DEFINE_string('suite_name', 'benchmark_suite',
                    'The name of the benchmark suite (for reporting).')
flags.DEFINE_string('base', None,
                    'JSON report produced by the base benchmark run.')
flags.DEFINE_string('exp', None,
                    'JSON report produced by the experiment benchmark run.')
flags.DEFINE_string('output', 'reports.csv', 'CSV output')
flags.DEFINE_multi_string(
    'counters', None,
    'Counter names. Should match exactly the names used when running the'
    'benchmark.')

FLAGS = flags.FLAGS


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')
  with tf.io.gfile.GFile(FLAGS.base, 'r') as b:
    with tf.io.gfile.GFile(FLAGS.exp, 'r') as e:
      b = benchmark_report.BenchmarkReport(FLAGS.suite_name, json.load(b),
                                           FLAGS.counters)
      e = benchmark_report.BenchmarkReport(FLAGS.suite_name, json.load(e),
                                           FLAGS.counters)
      comparison = benchmark_report.BenchmarkComparison(b, e)
      summary = comparison.summarize()
  with tf.io.gfile.GFile(FLAGS.output, 'w+') as o:
    co = csv.writer(o)
    # Pylint suggest using items for some reason, but the key is still needed
    # pylint: disable=consider-using-dict-items
    for bm in summary:
      for c in summary[bm]:
        co.writerow([bm, c] + list(summary[bm][c]))


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""RegAlloc priority prediction training config."""

import gin
import tensorflow as tf
from tf_agents.specs import tensor_spec
from tf_agents.trajectories import time_step
from compiler_opt.rl import feature_ops


@gin.configurable()
def get_regalloc_signature_spec():
  observation_spec = dict(
      (key, tf.TensorSpec(dtype=tf.int64, shape=(), name=key))
      for key in ('li_size', 'stage'))
  observation_spec['weight'] = tf.TensorSpec(
      dtype=tf.float32, shape=(), name='weight')

  reward_spec = tf.TensorSpec(dtype=tf.float32, shape=(), name='reward')
  time_step_spec = time_step.time_step_spec(observation_spec, reward_spec)

  action_spec = tensor_spec.TensorSpec(
      dtype=tf.float32, shape=(), name='priority')

  return time_step_spec, action_spec


@gin.configurable
def get_observation_processing_layer_creator():

  def observation_processing_layer(obs_spec):
    """Creates the layer to process observation given obs_spec."""

    if obs_spec.name in ('li_size', 'stage', 'weight'):
      return tf.keras.layers.Lambda(feature_ops.identity_fn)

    # Make sure all features have a preprocessing function.
    raise KeyError('Missing preprocessing function for some feature.')

  return observation_processing_layer
